# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import asyncio
import sqlalchemy

from asyncmy import Connection, Pool, create_pool, cursors, errors

from sqlalchemy.sql import ClauseElement, Select, Insert, Update, Delete

from .base import Utils

from ..interface import AsyncContextManager


MYSQL_POLL_WATER_LEVEL_WARNING_LINE = 0x10

MYSQL_EXECUTE_ARGUMENTS_TYPE = typing.Union[None, typing.Tuple, typing.List, typing.Dict]


# 数据库只读限制异常
class MySQLReadOnlyError(errors.MySQLError):
    pass


class MySQLClient(AsyncContextManager):
    """MySQL客户端对象，使用with进行上下文管理

    将连接委托给客户端对象管理，提高了整体连接的使用率

    """

    def __init__(self, pool: 'MySQLPool'):

        self._pool: MySQLPool = pool

        self._connection: typing.Optional[Connection] = None

        self._readonly: bool = pool.readonly

        self._lock: asyncio.Lock = asyncio.Lock()

    @property
    def readonly(self):

        return self._readonly

    @property
    def insert_id(self) -> int:

        if self._connection is not None:
            return self._connection.insert_id()

    async def _context_release(self):

        await self._close_connection()

    async def release(self):

        async with self._lock:
            await self._close_connection()

    async def _get_connection(self):

        if self._connection is None:
            self._connection = await self._pool.acquire()

        return self._connection

    async def _close_connection(self, discard=False):

        if self._connection is not None:

            if discard is True:
                await self._connection.ensure_closed()

            await self._pool.release(self._connection)

    async def _execute(self, query: str, args: MYSQL_EXECUTE_ARGUMENTS_TYPE = None) -> cursors.Cursor:

        try:

            connection = await self._get_connection()

            cursor = connection.cursor(cursor=self._pool.cursor_class)
            await cursor.execute(query, args)

            return cursor

        except Exception as err:

            await self._close_connection(True)

            raise err

    async def execute(self, query: str, args: MYSQL_EXECUTE_ARGUMENTS_TYPE = None) -> cursors.Cursor:

        async with self._lock:
            return await self._execute(query, args)

    @staticmethod
    def _clause_compile(query: ClauseElement) -> str:

        return query.compile(compile_kwargs={r'literal_binds': True}).string

    async def select(self, query: Select) -> typing.List[typing.Any]:

        async with self._lock:

            cursor = await self._execute(self._clause_compile(query))

            if cursor is not None:

                result = await cursor.fetchall()

                await cursor.close()

                return result

    async def find(self, query: Select) -> typing.Any:

        async with self._lock:

            cursor = await self._execute(self._clause_compile(query.limit(1)))

            if cursor is not None:

                result = await cursor.fetchone()

                await cursor.close()

                return result

    async def count(self, column: sqlalchemy.Column, where_clause: sqlalchemy.ColumnElement[bool] = None) -> int:

        _select = sqlalchemy.select([sqlalchemy.func.count(column).label(r'tbl_row_count')])

        if where_clause is not None:
            _select = _select.where(where_clause)

        cursor = await self.execute(self._clause_compile(_select))

        if cursor is not None:

            result = await cursor.fetchone()

            await cursor.close()

            return result[r'tbl_row_count']

    async def insert(self, query: Insert) -> int:

        if self._readonly:
            raise MySQLReadOnlyError()

        async with self._lock:

            cursor = await self._execute(self._clause_compile(query))

            if cursor is not None:

                result = self.insert_id

                await cursor.close()

                return result

    async def update(self, query: Update) -> int:

        if self._readonly:
            raise MySQLReadOnlyError()

        async with self._lock:

            cursor = await self._execute(self._clause_compile(query))

            if cursor is not None:

                result = cursor.rowcount

                await cursor.close()

                return result

    async def delete(self, query: Delete) -> int:

        if self._readonly:
            raise MySQLReadOnlyError()

        async with self._lock:

            cursor = await self._execute(self._clause_compile(query))

            if cursor is not None:

                result = cursor.rowcount

                await cursor.close()

                return result


class MySQLTransaction(MySQLClient):
    """MySQL客户端事务对象，使用with进行上下文管理

    将连接委托给客户端对象管理，提高了整体连接的使用率

    """

    def __init__(self, pool: 'MySQLPool'):

        super().__init__(pool)

        self._transaction_begin: bool = False

    async def _get_connection(self) -> Connection:

        if self._readonly is True:
            raise MySQLReadOnlyError()

        self._transaction_begin = True

        connection = await super()._get_connection()
        await connection.begin()

        return connection

    async def _context_release(self):

        await self.rollback()

    async def release(self):

        await self.rollback()

    async def commit(self):

        async with self._lock:

            if self._transaction_begin is True:
                await self._connection.commit()

            await self._close_connection()

    async def rollback(self):

        async with self._lock:

            if self._transaction_begin is True:
                await self._connection.rollback()

            await self._close_connection()


class MySQLPool:
    """MySQL连接管理
    """

    def __init__(
            self, host: str, port: int, db: str, user: str, password: str,
            *, name: str = None, minsize: int = 8, maxsize: int = 32, echo: bool = False, pool_recycle: int = 21600,
            charset: str = r'utf8', autocommit: bool = True,
            cursor_class: cursors.Cursor = cursors.DictCursor, readonly: bool = False,
            **settings
    ):

        self._name: str = name if name is not None else (Utils.uuid1()[:8] + (r'_ro' if readonly else r'_rw'))
        self._pool: typing.Optional[Pool] = None

        self._cursor_class: cursors.Cursor = cursor_class
        self._readonly: bool = readonly

        self._settings: typing.Dict[str, typing.Any] = settings

        self._settings[r'host'] = host
        self._settings[r'port'] = port
        self._settings[r'db'] = db

        self._settings[r'user'] = user
        self._settings[r'password'] = password

        self._settings[r'minsize'] = minsize
        self._settings[r'maxsize'] = maxsize

        self._settings[r'echo'] = echo
        self._settings[r'pool_recycle'] = pool_recycle
        self._settings[r'charset'] = charset
        self._settings[r'autocommit'] = autocommit

    @property
    def name(self) -> str:

        return self._name

    @property
    def readonly(self) -> bool:

        return self._readonly

    @property
    def cursor_class(self) -> cursors.Cursor:

        return self._cursor_class

    async def open(self):

        self._pool = await create_pool(**self._settings)

        Utils.log.info(
            f"MySQL [{self._settings[r'host']}:{self._settings[r'port']}] {self._settings[r'db']}"
            f" ({self._name}) initialized: {self._pool.size}/{self._pool.maxsize}"
        )

    async def close(self):

        if self._pool is not None:

            self._pool.close()
            await self._pool.wait_closed()

            self._pool = None

    def _echo_pool_info(self):

        global MYSQL_POLL_WATER_LEVEL_WARNING_LINE

        if (self._pool.maxsize - self._pool.size + self._pool.freesize) > MYSQL_POLL_WATER_LEVEL_WARNING_LINE:
            Utils.log.debug(
                f'MySQL connection pool info ({self._name}): '
                f'{self._pool.freesize}({self._pool.size}/{self._pool.maxsize})'
            )
        else:
            Utils.log.warning(
                f'MySQL connection pool not enough ({self._name}): '
                f'{self._pool.freesize}({self._pool.size}/{self._pool.maxsize})'
            )

    async def acquire(self) -> Connection:

        self._echo_pool_info()

        return await self._pool.acquire()

    async def release(self, connection: Connection):

        if self._pool is not None:
            self._pool.release(connection)
        else:
            await connection.close()

    def get_client(self) -> MySQLClient:

        return MySQLClient(self)

    def get_transaction(self) -> MySQLTransaction:

        if self._readonly:
            raise MySQLReadOnlyError()

        return MySQLTransaction(self)


class MySQLDelegate:
    """MySQL功能组件
    """

    def __init__(self):

        self._mysql_rw_pool: typing.Optional[MySQLPool] = None
        self._mysql_ro_pool: typing.Optional[MySQLPool] = None

    @property
    def mysql_rw_pool(self) -> MySQLPool:

        return self._mysql_rw_pool

    @property
    def mysql_ro_pool(self) -> MySQLPool:

        return self._mysql_ro_pool

    async def init_mysql_rw(self, *args, **kwargs):

        self._mysql_rw_pool = MySQLPool(*args, readonly=False, **kwargs)
        await self._mysql_rw_pool.open()

    async def init_mysql_ro(self, *args, **kwargs):

        self._mysql_ro_pool = MySQLPool(*args, readonly=True, **kwargs)
        await self._mysql_ro_pool.open()

    async def close_mysql(self):

        if self._mysql_rw_pool is not None:
            await self._mysql_rw_pool.close()

        if self._mysql_ro_pool is not None:
            await self._mysql_ro_pool.close()

    def get_mysql_client(self, readonly=False) -> MySQLClient:

        if readonly:
            if self._mysql_ro_pool:
                return self._mysql_ro_pool.get_client()
            else:
                return self._mysql_rw_pool.get_client()
        else:
            return self._mysql_rw_pool.get_client()

    def get_mysql_transaction(self) -> MySQLTransaction:

        return self._mysql_rw_pool.get_transaction()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import asyncio
import contextlib

from abc import abstractmethod

from .base import install_uvloop, Utils
from .future import WaitForever
from .socket import UnixSocketServer, UnixSocketClient, DEFAULT_LIMIT, DEFAULT_UNIX_SOCKET_ENDPOINT

from ..interface import RunnableInterface
from ..process import Daemon


class MainProcess(Daemon):

    def __init__(
            self, target: typing.Callable, sub_process_num: int, *,
            cpu_affinity: bool = False, join_timeout: int = 10,
            **kwargs
    ):

        super().__init__(
            target, sub_process_num,
            cpu_affinity=cpu_affinity, join_timeout=join_timeout,
            **kwargs
        )

    async def initialize(self):
        pass

    async def release(self):
        pass

    async def _execute(self):

        self._fill_process()

        while self._check_process():
            await Utils.sleep(1)

    def run(self):

        Utils.print_slogan()

        install_uvloop()

        loop = asyncio.get_event_loop()

        loop.run_until_complete(self.initialize())
        loop.run_until_complete(self._execute())
        loop.run_until_complete(self.release())


class SubProcess(RunnableInterface):

    @classmethod
    def create(cls, *args, **kwargs):
        cls(*args, **kwargs).run()

    def __init__(self, process_num: int, *args, **kwargs):

        self._process_id: int = Utils.getpid()
        self._process_num: int = process_num

        self._blocker: typing.Optional[WaitForever] = None

    def close(self, *_):

        if self._blocker and not self._blocker.cancelled():
            self._blocker.cancel()

    async def initialize(self):
        pass

    async def release(self):
        pass

    async def _execute(self):

        with contextlib.suppress(asyncio.CancelledError):
            self._blocker = WaitForever()
            await self._blocker

    def run(self):

        Utils.log.success(f'Started worker process [id:{self._process_id} num:{self._process_num}]')

        install_uvloop()

        loop = asyncio.get_event_loop()

        loop.run_until_complete(self.initialize())
        loop.run_until_complete(self._execute())
        loop.run_until_complete(self.release())

        Utils.log.success(f'Stopped worker process [id:{self._process_id} num:{self._process_num}]')


class MainProcessWithIPC(MainProcess):

    def __init__(
            self, target: typing.Callable, sub_process_num: int, *,
            cpu_affinity: bool = False, join_timeout: int = 10,
            unix_socket_path: str = DEFAULT_UNIX_SOCKET_ENDPOINT, unix_socket_limit: int = DEFAULT_LIMIT,
            **kwargs
    ):

        super().__init__(
            target, sub_process_num,
            cpu_affinity=cpu_affinity, join_timeout=join_timeout,
            unix_socket_path=unix_socket_path,
            **kwargs
        )

        self._socket_server: UnixSocketServer = UnixSocketServer(
            self._client_connected_cb, unix_socket_path, unix_socket_limit
        )

    @abstractmethod
    async def _client_connected_cb(
            self,
            reader: 'asyncio.streams.StreamReader',
            writer: 'asyncio.streams.StreamWriter'
    ):
        """
        连接成功后的回调函数
        """

    async def initialize(self):
        await self._socket_server.open()

    async def release(self):
        await self._socket_server.close()


class SubProcessWithIPC(SubProcess):

    def __init__(self, process_num: int, unix_socket_path: str):

        super().__init__(process_num)

        self._socket_client: UnixSocketClient = UnixSocketClient(unix_socket_path)

    async def initialize(self):
        await self._socket_client.open()

    async def release(self):
        await self._socket_client.close()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import re
import enum
import copy
import struct
import threading
import itertools

from abc import ABCMeta, abstractmethod
from io import BytesIO


class Enum(enum.Enum):

    @classmethod
    def items(cls) -> typing.Iterable:
        return cls.to_dict().items()

    @classmethod
    def keys(cls) -> typing.Iterable:
        return cls.to_dict().keys()

    @classmethod
    def values(cls) -> typing.Iterable:
        return cls.to_dict().values()

    @classmethod
    def to_dict(cls) -> typing.Dict:

        if not hasattr(cls, r'_to_dict'):
            setattr(cls, r'_to_dict', {item.name: item.value for item in cls})

        return getattr(cls, r'_to_dict')

    @classmethod
    def to_keys_dict(cls) -> typing.Dict:

        if not hasattr(cls, r'_to_keys_dict'):
            setattr(cls, r'_to_keys_dict', {item.name: item for item in cls})

        return getattr(cls, r'_to_keys_dict')

    @classmethod
    def to_values_dict(cls) -> typing.Dict:

        if not hasattr(cls, r'_to_values_dict'):
            setattr(cls, r'_to_values_dict', {item.value: item for item in cls})

        return getattr(cls, r'_to_values_dict')

    @classmethod
    def has_key(cls, key) -> bool:
        return key in cls.to_keys_dict()

    @classmethod
    def has_value(cls, value) -> bool:
        return value in cls.to_values_dict()


class IntEnum(int, Enum): pass


RoundRobinMemberType = typing.TypeVar('RoundRobinMemberType')

class RoundRobin(typing.Generic[RoundRobinMemberType]):

    def __init__(self, members: typing.Optional[typing.Dict[str, RoundRobinMemberType]] = None):

        self._members: typing.Dict[str, RoundRobinMemberType] = members if members else {}
        self._iterator: typing.Iterator = itertools.cycle(self._members.items())

    def __len__(self) -> int:
        return len(self._members)

    def __contains__(self, key) -> bool:
        return key in self._members

    def append(self, name: str, member: RoundRobinMemberType) -> typing.Optional[RoundRobinMemberType]:

        if name in self._members:
            _member = self._members.pop(name)
        else:
            _member = None

        self._members[name] = member
        self._iterator = itertools.cycle(self._members.items())

        return _member

    def clear(self, names: typing.Optional[typing.List[str]] = None) -> typing.Dict[str, RoundRobinMemberType]:

        members = {}

        if names is None:

            members.update(self._members)
            self._members.clear()

        else:

            for _name in names:
                if _name in self._members:
                    members[_name] = self._members.pop(_name)

        self._iterator = itertools.cycle(self._members.items())

        return members

    def reset(self, members: typing.Optional[typing.Dict[str, RoundRobinMemberType]] = None) -> typing.Dict[str, RoundRobinMemberType]:

        _members = members if members else {}

        _members, self._members = self._members, _members
        self._iterator = itertools.cycle(self._members.items())

        return _members

    def get(self) -> typing.Tuple[str, RoundRobinMemberType]:
        return next(self._iterator)

class RoundRobinSimple(typing.Generic[RoundRobinMemberType]):

    def __init__(self, members: typing.Optional[typing.List[RoundRobinMemberType]] = None):
        self._members: typing.List[RoundRobinMemberType] = members if members else []
        self._iterator: typing.Iterator = itertools.cycle(self._members)

    def __len__(self) -> int:
        return len(self._members)

    def append(self, member: RoundRobinMemberType):
        self._members.append(member)
        self._iterator = itertools.cycle(self._members)

    def clear(self) -> typing.List[RoundRobinMemberType]:

        members = copy.copy(self._members)

        self._members.clear()
        self._iterator = itertools.cycle(self._members)

        return members

    def get(self) -> RoundRobinMemberType:
        return next(self._iterator)


class ThreadList(threading.local):
    """多线程安全的列表
    """

    __slots__ = [r'data']

    def __init__(self):

        self.data: typing.List = []


class ThreadDict(threading.local):
    """多线程安全的字典
    """

    __slots__ = [r'data']

    def __init__(self):

        self.data: typing.Dict = {}


class ByteArrayAbstract(metaclass=ABCMeta):
    """ByteArray抽象类
    """

    NETWORK = r'!'
    NATIVE = r'='
    NATIVE_ALIGNMENT = r'@'
    LITTLE_ENDIAN = r'<'
    BIG_ENDIAN = r'>'

    def __init__(self):

        self._endian: str = self.NETWORK

    def get_endian(self) -> str:

        return self._endian

    def set_endian(self, val: str):

        self._endian = val

    @abstractmethod
    def read(self, size) -> typing.Any:
        """
        读取数据
        """

    @abstractmethod
    def write(self, buffer: typing.Any):
        """
        定稿数据
        """

    def read_pad_byte(self, size: int) -> typing.Tuple:

        return struct.unpack(f'{self._endian}{size}x', self.read(size))

    def write_pad_byte(self, val: typing.Tuple):

        self.write(struct.pack(f'{self._endian}{val}x'))

    def read_char(self) -> int:

        return struct.unpack(f'{self._endian}c', self.read(1))[0]

    def write_char(self, val: int):

        self.write(struct.pack(f'{self._endian}c', val))

    def read_signed_char(self) -> int:

        return struct.unpack(f'{self._endian}b', self.read(1))[0]

    def write_signed_char(self, val: int):

        self.write(struct.pack(f'{self._endian}b', val))

    def read_unsigned_char(self) -> int:

        return struct.unpack(f'{self._endian}B', self.read(1))[0]

    def write_unsigned_char(self, val: int):

        self.write(struct.pack(f'{self._endian}B', val))

    def read_bool(self) -> bool:

        return struct.unpack(f'{self._endian}?', self.read(1))[0]

    def write_bool(self, val: bool):

        self.write(struct.pack(f'{self._endian}?', val))

    def read_short(self) -> int:

        return struct.unpack(f'{self._endian}h', self.read(2))[0]

    def write_short(self, val: int):

        self.write(struct.pack(f'{self._endian}h', val))

    def read_unsigned_short(self) -> int:

        return struct.unpack(f'{self._endian}H', self.read(2))[0]

    def write_unsigned_short(self, val: int):

        self.write(struct.pack(f'{self._endian}H', val))

    def read_int(self) -> int:

        return struct.unpack(f'{self._endian}i', self.read(4))[0]

    def write_int(self, val: int):

        self.write(struct.pack(f'{self._endian}i', val))

    def read_unsigned_int(self) -> int:

        return struct.unpack(f'{self._endian}I', self.read(4))[0]

    def write_unsigned_int(self, val: int):

        self.write(struct.pack(f'{self._endian}I', val))

    def read_long(self) -> int:

        return struct.unpack(f'{self._endian}l', self.read(4))[0]

    def write_long(self, val: int):

        self.write(struct.pack(f'{self._endian}l', val))

    def read_unsigned_long(self) -> int:

        return struct.unpack(f'{self._endian}L', self.read(4))[0]

    def write_unsigned_long(self, val: int):

        self.write(struct.pack(f'{self._endian}L', val))

    def read_long_long(self) -> int:

        return struct.unpack(f'{self._endian}q', self.read(8))[0]

    def write_long_long(self, val: int):

        self.write(struct.pack(f'{self._endian}q', val))

    def read_unsigned_long_long(self) -> int:

        return struct.unpack(f'{self._endian}Q', self.read(8))[0]

    def write_unsigned_long_long(self, val: int):

        self.write(struct.pack(f'{self._endian}Q', val))

    def read_float(self) -> float:

        return struct.unpack(f'{self._endian}f', self.read(4))[0]

    def write_float(self, val):

        self.write(struct.pack(f'{self._endian}f', val))

    def read_double(self) -> float:

        return struct.unpack(f'{self._endian}d', self.read(8))[0]

    def write_double(self, val: float):

        self.write(struct.pack(f'{self._endian}d', val))

    def read_bytes(self, size: int) -> bytes:

        return struct.unpack(f'{self._endian}{size}s', self.read(size))[0]

    def write_bytes(self, val: bytes):

        self.write(struct.pack(f'{self._endian}{len(val)}s', val))

    def read_string(self, size: int):

        return self.read_bytes(size).decode()

    def write_string(self, val: str):

        self.write_bytes(val.encode())

    def read_pascal_bytes(self, size: int):

        return struct.unpack(f'{self._endian}{size}p', self.read(size))[0]

    def write_pascal_bytes(self, val: bytes):

        self.write(struct.pack(f'{self._endian}{len(val)}p', val))

    def read_pascal_string(self, size: int):

        return self.read_pascal_bytes(size).decode()

    def write_pascal_string(self, val: str):

        self.write_pascal_bytes(val.encode())


class ByteArray(BytesIO, ByteArrayAbstract):
    """扩展的BytesIO类
    """

    NETWORK = r'!'
    NATIVE = r'='
    NATIVE_ALIGNMENT = r'@'
    LITTLE_ENDIAN = r'<'
    BIG_ENDIAN = r'>'

    def __init__(self, initial_bytes: typing.Optional[bytes] = None):

        BytesIO.__init__(self, initial_bytes)
        ByteArrayAbstract.__init__(self)


class KeyLowerDict(dict):

    _PATTERN = re.compile(r'(?<=[a-z])([A-Z])')

    def __init__(self, _dict: typing.Dict[str, typing.Any]):

        super().__init__(
            {
                KeyLowerDict._PATTERN.sub(r'_\1', key).lower(): KeyLowerDict(val) if isinstance(val, dict) else val
                for key, val in _dict.items()
            }
        )



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import pytz
import typing
import asyncio
import logging

from datetime import datetime
from abc import abstractmethod
from coredis import PureToken

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger
from apscheduler.job import Job

from ..interface import TaskInterface

from .base import Utils
from .redis import RedisPool, RedisClusterPool
from ..error import catch_error


DEFAULT_SCHEDULER_TIMEZONE = pytz.timezone(r'Asia/Shanghai')

DEFAULT_SCHEDULER_CONFIG = {
    r'job_defaults': {
        r'coalesce': False,
        r'max_instances': 1,
        r'misfire_grace_time': 10
    },
    r'timezone': DEFAULT_SCHEDULER_TIMEZONE
}

logging.getLogger(r'apscheduler').setLevel(logging.ERROR)


class TaskAbstract(TaskInterface):
    """任务基类
    """

    def __init__(self, scheduler: typing.Optional[AsyncIOScheduler] = None):
        self._scheduler: AsyncIOScheduler = AsyncIOScheduler(
            **DEFAULT_SCHEDULER_CONFIG
        ) if scheduler is None else scheduler

    @property
    def scheduler(self) -> AsyncIOScheduler:
        return self._scheduler

    def start(self):
        if not self._scheduler.running:
            self._scheduler.start()

    def stop(self):
        if self._scheduler.running:
            self._scheduler.shutdown()

    def is_running(self):
        return self._scheduler.running

    @abstractmethod
    def add_job(self, *args, **kwargs) -> Job:
        """
        添加任务
        """

    def remove_job(self, job_id):
        self._scheduler.remove_job(job_id)

    def remove_all_jobs(self):
        self._scheduler.remove_all_jobs()


class IntervalTask(TaskAbstract):
    """间隔任务类
    """

    @classmethod
    def create(cls, interval: int, func: typing.Callable, *func_args, **func_kwargs) -> 'IntervalTask':

        inst = cls()

        inst.add_job(interval, func, args=func_args, kwargs=func_kwargs)

        return inst

    def add_job(self, interval: int, func: typing.Callable, *args, **kwargs) -> Job:
        return self._scheduler.add_job(
            func, r'interval', *args, seconds=interval, next_run_time=datetime.now(), **kwargs
        )


class CronTask(TaskAbstract):
    """定时任务类
    """

    @classmethod
    def create(cls, crontab: str, func: typing.Callable, *func_args, **func_kwargs) -> 'CronTask':

        inst = cls()

        inst.add_job(crontab, func, args=func_args, kwargs=func_kwargs)

        return inst

    def add_job(self, crontab: str, func: typing.Callable, *args, **kwargs) -> Job:
        return self._scheduler.add_job(
            func, CronTrigger.from_crontab(crontab, DEFAULT_SCHEDULER_TIMEZONE), *args, **kwargs
        )


class DCSCronTask(TaskInterface):

    def __init__(
            self,
            redis_client: typing.Union[RedisPool, RedisClusterPool],
            name: str,
            crontab: str,
            func: typing.Callable,
            *args, **kwargs
    ):

        self._redis_client: typing.Union[RedisPool, RedisClusterPool] = redis_client

        self._name: str = name
        self._task_func: typing.Callable = func
        self._cron_task: CronTask = CronTask.create(crontab, self._do_job, *args, **kwargs)

    async def _do_job(self, *args, **kwargs):

        key = self._redis_client.get_safe_key(
            r'dcs_cron', self._name, Utils.stamp2time(format_type=r'%Y%m%d%H%M')
        )

        if await self._redis_client.set(key, self._name, condition=PureToken.NX, ex=3600):

            Utils.log.info(f'dcs cron task start: {self._name}')

            if Utils.is_coroutine_function(self._task_func):
                await self._task_func(*args, **kwargs)
            else:
                self._task_func(*args, **kwargs)

            Utils.log.info(f'dcs cron task finish: {self._name}')

        else:

            Utils.log.debug(f'dcs cron task idle: {self._name}')

    def start(self):
        self._cron_task.start()
        Utils.log.info(f'dcs cron task init: {self._name}')

    def stop(self):
        self._cron_task.stop()

    def is_running(self) -> bool:
        return self._cron_task.is_running()


class MultiTasks:
    """多任务并发管理器

    提供协程的多任务并发的解决方案, 通过参数tasks_num控制队列长度

    tasks = MultiTasks()
    tasks.append(func1())
    tasks.append(func2())
    ...
    tasks.append(funcN())
    await tasks

    """

    def __init__(self, tasks_num: int = 10) -> None:

        self._semaphore: asyncio.Semaphore = asyncio.Semaphore(max(1, tasks_num))

        self._coroutines: typing.List[typing.Coroutine] = []

    def __await__(self) -> typing.List[typing.Any]:

        result = None

        if len(self._coroutines) > 0:

            tasks = [self._do_task(_task) for _task in self._coroutines]

            result = yield from asyncio.gather(*tasks).__await__()

            self._coroutines.clear()

        return result

    async def _do_task(self, coroutine: typing.Coroutine) -> typing.Any:

        async with self._semaphore:
            return await coroutine

    def append(self, coroutine: typing.Coroutine):
        return self._coroutines.append(coroutine)

    def extend(self, coroutines: typing.List[typing.Coroutine]):
        return self._coroutines.extend(coroutines)


class RateLimiter:
    """流量控制器，用于对计算资源的保护
    进入队列的任务，如果触发限流行为会通过在Future上引发CancelledError传递出来
    """

    def __init__(self, task_limit: int, wait_limit: int = 256, timeout: int = 60):

        self._semaphore: asyncio.Semaphore = asyncio.Semaphore(max(1, task_limit))

        self._task_queue: asyncio.Queue[
            typing.Tuple[typing.Callable, typing.Optional[asyncio.Future], float]
        ] = asyncio.Queue(wait_limit)

        self._timeout: int = max(1, timeout)

        self._consume_task: asyncio.Task = asyncio.create_task(self._do_consume_task())

    async def _put_task_queue(
            self, func: typing.Callable, func_args: typing.Tuple, func_kwargs: typing.Dict,
            future: typing.Optional[asyncio.Future] = None
    ) -> bool:

        if not Utils.is_coroutine_function(func):
            raise TypeError('must be a coroutine function')

        await asyncio.wait_for(
            self._task_queue.put(
                (Utils.func_partial(func, *func_args, **func_kwargs), future, Utils.loop_time())
            ),
            self._timeout
        )

        return True

    def _create_task(self, func: typing.Callable, future: typing.Optional[asyncio.Future]):

        task = asyncio.create_task(func())
        task.add_done_callback(lambda _ : self._task_queue.task_done())
        task.add_done_callback(lambda _ : self._semaphore.release())

        if future is not None:
            task.add_done_callback(lambda _: future.set_result(task.result()))

    async def _do_consume_task(self):

        while True:

            future = None

            try:

                await self._semaphore.acquire()

                func, future, join_time = await self._task_queue.get()

                if (Utils.loop_time() - join_time) > self._timeout:
                    raise TimeoutError(f'timeout')

                self._create_task(func, future)

            except Exception as err:

                Utils.log.error(f'rate limit error: {str(err)}')

                if future is not None:
                    future.set_exception(err)

    def size(self) -> int:
        return self._task_queue.qsize()

    def close(self):
        self._consume_task.cancel()

    async def join(self):
        await self._task_queue.join()

    async def append(self, func: typing.Callable, *args, **kwargs) -> bool:

        result = False

        try:
            result = await self._put_task_queue(func, args, kwargs)
        except Exception as err:
            Utils.log.error(str(err))

        return result

    async def call(self, func: typing.Callable, *args, **kwargs) -> typing.Any:

        future = asyncio.Future()

        try:
            await self._put_task_queue(func, args, kwargs, future)
        except Exception as err:
            Utils.log.error(str(err))
            future.set_exception(err)

        return await future



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import sys

os.chdir(os.path.dirname(__file__))
sys.path.insert(0, os.path.abspath(r'../../'))

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from hagworm.frame.fastapi.base import uvicorn_run, create_fastapi
from hagworm.extend.logging import LogFileRotator

from setting import Config
from service import DataSource
from controller import router


def app() -> FastAPI:

    fastapi = create_fastapi(
        log_level=Config.LogLevel,
        log_file_path=Config.LogFilePath,
        log_file_rotation=LogFileRotator.make(Config.LogFileSplitSize, Config.LogFileSplitTime),
        log_file_retention=Config.LogFileBackups,
        debug=Config.Debug,
        routes=router.routes,
        on_startup=[DataSource.initialize],
        on_shutdown=[DataSource.release],
    )

    fastapi.add_middleware(
        CORSMiddleware,
        allow_origins=Config.AllowOrigins,
        allow_methods=Config.AllowMethods,
        allow_headers=Config.AllowHeaders,
        allow_credentials=Config.AllowCredentials,
    )

    return fastapi


if __name__ == r'__main__':
    uvicorn_run(app, port=Config.Port)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing

import asyncio
import inspect
import aio_pika
import async_timeout

from .publish import RabbitMQProducerForExchangePool
from .consume import RabbitMQConsumerForExchange

from ...extend.error import catch_error
from ...extend.asyncio.base import Utils


STREAM_EOF: bytes = b'\0\0'


class RpcServerError(Exception): pass


def get_routing_key(server_name: typing.Optional[str]) -> str:
    return f'rpc_server.{server_name}' if server_name else r'rpc_server'


class RpcServer:

    def __init__(
            self, pool_size: int, connection: aio_pika.RobustConnection, exchange_name: str,
            *, server_name: typing.Optional[str] = None, message_ttl: float = 60, max_priority: int = 10
    ):

        self._rpc_function: typing.Dict[str, typing.Callable] = {}

        self._queue_name: str = get_routing_key(server_name)
        self._exchange_name: str = exchange_name

        self._queue_config: typing.Dict[str, typing.Any] = {
            r'arguments': {
                r'x-message-ttl': round(message_ttl * 1000),
                r'x-max-priority': max_priority,
            }
        }

        self._producer: RabbitMQProducerForExchangePool = RabbitMQProducerForExchangePool(
            pool_size, connection, self._exchange_name
        )

        self._consumer: RabbitMQConsumerForExchange = RabbitMQConsumerForExchange(
            connection, self._queue_name, self._exchange_name
        )

    async def open(self):

        self._rpc_function[r'ping'] = lambda: r'pong'

        await self._producer.open(
            exchange_type=aio_pika.ExchangeType.TOPIC
        )

        await self._consumer.open(
            consume_func=self._consume_message,
            consume_no_ack=False,
            queue_config=self._queue_config,
            channel_qos_config={r'prefetch_count': self._producer.size}
        )

    async def close(self):

        await self._producer.close()
        await self._consumer.close()

        self._rpc_function.clear()

    async def _reply_message(self, incoming: aio_pika.IncomingMessage, func_name: str, data: typing.Any):

        await self._producer.publish(
            aio_pika.Message(
                body=Utils.msgpack_encode(
                    {
                        r'name': func_name,
                        r'data': data,
                    }
                ),
                correlation_id=incoming.correlation_id
            ),
            routing_key=incoming.reply_to
        )

    async def _reply_error(self, incoming: aio_pika.IncomingMessage, func_name: str, error: str):

        await self._producer.publish(
            aio_pika.Message(
                body=Utils.msgpack_encode(
                    {
                        r'name': func_name,
                        r'error': error,
                    }
                ),
                correlation_id=incoming.correlation_id
            ),
            routing_key=incoming.reply_to
        )

    async def _reply_eof(self, incoming: aio_pika.IncomingMessage):

        await self._producer.publish(
            aio_pika.Message(
                body=STREAM_EOF,
                correlation_id=incoming.correlation_id
            ),
            routing_key=incoming.reply_to
        )

    async def _consume_message(self, message: aio_pika.IncomingMessage):

        with catch_error():

            _message = Utils.msgpack_decode(message.body)

            Utils.log.info(f'rpc client request: {_message}')

            _func_name = _message.get(r'name')
            _func_args = _message.get(r'args')
            _func_kwargs = _message.get(r'kwargs')

            if _func_name not in self._rpc_function:
                raise Exception(f'function {_func_name} not found')

            try:

                result = self._rpc_function[_func_name](*_func_args, **_func_kwargs)

                if inspect.iscoroutine(result):
                    await self._reply_message(message, _func_name, await result)
                elif inspect.isasyncgen(result):
                    async for _item in result:
                        await self._reply_message(message, _func_name, _item)
                    else:
                        await self._reply_eof(message)
                else:
                    raise Exception(f'function {_func_name} not supported')

            except Exception as err:
                await self._reply_error(message, _func_name, str(err))
                raise err

        await message.ack()

    def register(self, name: str, func: typing.Callable):

        self._rpc_function[name] = func

        Utils.log.info(f'rpc server register {name} {func}')


class RpcClient:

    def __init__(
            self, pool_size: int, connection: aio_pika.RobustConnection,
            exchange_name: str, message_ttl: float = 60
    ):

        self._unary_response: typing.Dict[str, asyncio.Future] = {}
        self._stream_response: typing.Dict[str, asyncio.Queue] = {}

        self._queue_name: str = f'rpc_client.{Utils.uuid1()}'
        self._exchange_name: str = exchange_name

        self._queue_config: typing.Dict[str, typing.Any] = {
            r'exclusive': True,
            r'arguments': {r'x-message-ttl': round(message_ttl * 1000)}
        }

        self._producer: RabbitMQProducerForExchangePool = RabbitMQProducerForExchangePool(
            pool_size, connection, self._exchange_name
        )

        self._consumer: RabbitMQConsumerForExchange = RabbitMQConsumerForExchange(
            connection, self._queue_name, self._exchange_name
        )

    async def open(self):

        await self._producer.open(
            exchange_type=aio_pika.ExchangeType.TOPIC
        )

        await self._consumer.open(
            consume_func=self._consume_message,
            consume_no_ack=False,
            queue_config=self._queue_config,
            channel_qos_config={r'prefetch_count': 1}
        )

    async def close(self):

        await self._producer.close()
        await self._consumer.close()

    async def _send_message(self, correlation_id: str, priority: int, server_name: str, message: typing.Any):

        await self._producer.publish(
            aio_pika.Message(
                Utils.msgpack_encode(message),
                priority=priority,
                correlation_id=correlation_id,
                reply_to=self._consumer.queue_name,
            ),
            routing_key=get_routing_key(server_name),
        )

    async def _consume_message(self, message: aio_pika.IncomingMessage):

        with catch_error():

            message_body = message.body
            correlation_id = message.correlation_id

            if correlation_id in self._unary_response:
                self._unary_response[correlation_id].set_result(message_body)
            elif correlation_id in self._stream_response:
                self._stream_response[correlation_id].put_nowait(message_body)
            else:
                raise Exception(f'unexpected correlation_id: {correlation_id}')

        await message.ack()

    async def ping(self, priority: int = 0, server_name: typing.Optional[str] = None):
        return await self.call(r'ping', priority=priority, server_name=server_name)

    async def call(
            self, name: str, *,
            args: typing.Optional[typing.List] = None,
            kwargs: typing.Optional[typing.Dict] = None,
            priority: int = 0,
            server_name: typing.Optional[str] = None,
            request_timeout: float = 60
    ):

        correlation_id = Utils.uuid1()

        unary_response = self._unary_response[correlation_id] = asyncio.Future()

        try:

            async with async_timeout.timeout(request_timeout):

                await self._send_message(
                    correlation_id, priority, server_name,
                    {
                        r'name': name,
                        r'args': args if args is not None else [],
                        r'kwargs': kwargs if kwargs is not None else {},
                    }
                )

                response = Utils.msgpack_decode(await unary_response)

                if r'error' in response:
                    raise RpcServerError(response[r'error'])

                return response[r'data']

        except asyncio.TimeoutError as err:
            Utils.log.error(f'rpc call timeout: {name=} {args=} {kwargs=}')
            raise err

        finally:
            del self._unary_response[correlation_id]

    async def call_stream(
            self, name: str, *,
            args: typing.Optional[typing.List] = None,
            kwargs: typing.Optional[typing.Dict] = None,
            priority: int = 0,
            server_name: typing.Optional[str] = None,
            request_timeout: float = 3600
    ):

        correlation_id = Utils.uuid1()

        stream_response = self._stream_response[correlation_id] = asyncio.Queue()

        try:

            async with async_timeout.timeout(request_timeout):

                await self._send_message(
                    correlation_id, priority, server_name,
                    {
                        r'name': name,
                        r'args': args if args is not None else [],
                        r'kwargs': kwargs if kwargs is not None else {},
                    }
                )

                while (response := await stream_response.get()) != STREAM_EOF:

                    response = Utils.msgpack_decode(response)

                    if r'error' in response:
                        raise RpcServerError(response[r'error'])

                    yield response[r'data']

        except asyncio.TimeoutError as err:
            Utils.log.error(f'rpc call timeout: {name=} {args=} {kwargs=}')
            raise err

        finally:
            del self._stream_response[correlation_id]



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing

import qrcode

from io import BytesIO
from qrcode.image.svg import SvgPathImage


class QRCode:

    @staticmethod
    def make(data: typing.Any, **kwargs) -> bytes:

        code_img = BytesIO()

        qrcode.make(data, **kwargs).save(code_img)

        return code_img.getvalue()

    @staticmethod
    def make_svg(data: typing.Any, **kwargs) -> str:

        code_img = BytesIO()

        kwargs[r'image_factory'] = SvgPathImage

        qrcode.make(data, **kwargs).save(code_img)

        return code_img.getvalue().decode()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import time
import signal
import typing

from multiprocessing import Process, set_start_method
from multiprocessing.shared_memory import SharedMemory

from .base import Utils
from .interface import ContextManager, RunnableInterface
from .struct import ByteArrayAbstract


set_start_method(r'spawn', force=True)


def fork_processes() -> int:

    pid = os.fork()

    if pid == 0:
        Utils.urandom_seed()

    return pid


class Daemon(RunnableInterface):

    def __init__(
            self, target: typing.Callable, sub_process_num: int, *,
            cpu_affinity: bool = False, join_timeout: int = 10, **kwargs
    ):

        self._target: typing.Callable = target
        self._kwargs: typing.Dict = kwargs

        self._sub_process: typing.Set = set()
        self._sub_process_num: int = sub_process_num

        self._cpu_affinity: bool = cpu_affinity
        self._join_timeout: int = join_timeout

        signal.signal(signal.SIGINT, self._kill_process)
        signal.signal(signal.SIGTERM, self._kill_process)

        if self._cpu_affinity and sub_process_num > os.cpu_count():
            raise Exception(r'The number of processes exceeds the number of CPU cores')

    def _fill_process(self):

        for idx in range(self._sub_process_num - len(self._sub_process)):

            process = Process(target=self._target, args=(idx,), kwargs=self._kwargs)
            process.start()

            self._sub_process.add(process)

        if self._cpu_affinity:

            for _idx, _process in enumerate(self._sub_process):
                os.sched_setaffinity(_process.pid, [_idx])
                Utils.log.info(f'process {_process.pid} affinity: {os.sched_getaffinity(_process.pid)}')

    def _kill_process(self, *_):

        for process in self._sub_process:
            os.kill(process.ident, signal.SIGINT)

        for process in self._sub_process:
            process.join(self._join_timeout)
            process.kill()

    def _check_process(self):

        for process in self._sub_process.copy():

            if process.is_alive():
                continue

            self._sub_process.remove(process)
            Utils.log.warning(f'kill process {process.ident}')

        return len(self._sub_process) > 0

    def run(self):

        self._fill_process()

        while self._check_process():
            time.sleep(1)


class SharedByteArray(ByteArrayAbstract, ContextManager):

    def __init__(self, name: typing.Optional[str] = None, create: bool = False, size: int = 0):

        ByteArrayAbstract.__init__(self)

        self._shared_memory: SharedMemory = SharedMemory(name, create, size)
        self._create_mode: bool = create

    def _context_release(self):

        self.release()

    def release(self):

        self._shared_memory.close()

        if self._create_mode:
            self._shared_memory.unlink()

    def read(self, size: int) -> memoryview:
        return self._shared_memory.buf[:size]

    def write(self, buffer: memoryview):
        self._shared_memory.buf[:len(buffer)] = buffer



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import re


####################
# ASC可见字符验证

REGEX_ASC_VISIBLE: re.Pattern = re.compile(r'^[\x21-\x7e]$')


def asc_visible(val: str) -> bool:

    global REGEX_ASC_VISIBLE

    return True if REGEX_ASC_VISIBLE.match(val) else False


####################
# UUID验证

REGEX_UUID: re.Pattern = re.compile(r'^[0-9a-fA-F]{8}-([0-9a-fA-F]{4}-){3}[0-9a-fA-F]{12}$')


def uuid(val: str) -> bool:

    global REGEX_UUID

    return True if REGEX_UUID.match(val) else False


####################
# 邮箱验证

REGEX_EMAIL: re.Pattern = re.compile(r'^[a-z0-9]+[\._]?[a-z0-9]+[@]\w+[.]\w{2,3}$')


def email(val: str) -> bool:

    global REGEX_EMAIL

    return True if REGEX_EMAIL.match(val) else False


####################
# 域名验证

REGEX_DOMAIN: re.Pattern = re.compile(
    r'^(?:[a-zA-Z0-9]'  # First character of the domain
    r'(?:[a-zA-Z0-9-_]{0,61}[A-Za-z0-9])?\.)'  # Sub domain + hostname
    r'+[A-Za-z0-9][A-Za-z0-9-_]{0,61}'  # First 61 characters of the gTLD
    r'[A-Za-z]$'  # Last character of the gTLD
)


def domain(val: str) -> bool:

    global REGEX_DOMAIN

    return True if REGEX_DOMAIN.match(val) else False


####################
# URL验证

_IP_MIDDLE_OCTET = r'(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5]))'
_IP_LAST_OCTET = r'(?:\.(?:0|[1-9]\d?|1\d\d|2[0-4]\d|25[0-5]))'

REGEX_URL: re.Pattern = re.compile(
    r"^"
    # protocol identifier
    r"(?:(?:https?|ftp)://)"
    # user:pass authentication
    r"(?:[-a-z\u00a1-\uffff0-9._~%!$&'()*+,;=:]+"
    r"(?::[-a-z0-9._~%!$&'()*+,;=:]*)?@)?"
    r"(?:"
    r"(?P<private_ip>"
    # IP address exclusion
    # private & local networks
    r"(?:(?:10|127)" + _IP_MIDDLE_OCTET + r"{2}" + _IP_LAST_OCTET + r")|"
    r"(?:(?:169\.254|192\.168)" + _IP_MIDDLE_OCTET + _IP_LAST_OCTET + r")|"
    r"(?:172\.(?:1[6-9]|2\d|3[0-1])" + _IP_MIDDLE_OCTET + _IP_LAST_OCTET + r"))"
    r"|"
    # private & local hosts
    r"(?P<private_host>"
    r"(?:localhost))"
    r"|"
    # IP address dotted notation octets
    # excludes loopback network 0.0.0.0
    # excludes reserved space >= 224.0.0.0
    # excludes network & broadcast addresses
    # (first & last IP address of each class)
    r"(?P<public_ip>"
    r"(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])"
    r"" + _IP_MIDDLE_OCTET + r"{2}"
    r"" + _IP_LAST_OCTET + r")"
    r"|"
    # IPv6 RegEx from https://stackoverflow.com/a/17871737
    r"\[("
    # 1:2:3:4:5:6:7:8
    r"([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|"
    # 1::                              1:2:3:4:5:6:7::
    r"([0-9a-fA-F]{1,4}:){1,7}:|"
    # 1::8             1:2:3:4:5:6::8  1:2:3:4:5:6::8
    r"([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|"
    # 1::7:8           1:2:3:4:5::7:8  1:2:3:4:5::8
    r"([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|"
    # 1::6:7:8         1:2:3:4::6:7:8  1:2:3:4::8
    r"([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|"
    # 1::5:6:7:8       1:2:3::5:6:7:8  1:2:3::8
    r"([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|"
    # 1::4:5:6:7:8     1:2::4:5:6:7:8  1:2::8
    r"([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|"
    # 1::3:4:5:6:7:8   1::3:4:5:6:7:8  1::8
    r"[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|"
    # ::2:3:4:5:6:7:8  ::2:3:4:5:6:7:8 ::8       ::
    r":((:[0-9a-fA-F]{1,4}){1,7}|:)|"
    # fe80::7:8%eth0   fe80::7:8%1
    # (link-local IPv6 addresses with zone index)
    r"fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|"
    r"::(ffff(:0{1,4}){0,1}:){0,1}"
    r"((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}"
    # ::255.255.255.255   ::ffff:255.255.255.255  ::ffff:0:255.255.255.255
    # (IPv4-mapped IPv6 addresses and IPv4-translated addresses)
    r"(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|"
    r"([0-9a-fA-F]{1,4}:){1,4}:"
    r"((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}"
    # 2001:db8:3:4::192.0.2.33  64:ff9b::192.0.2.33
    # (IPv4-Embedded IPv6 Address)
    r"(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])"
    r")\]|"
    # host name
    r"(?:(?:(?:xn--)|[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]-?)*"
    r"[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]+)"
    # domain name
    r"(?:\.(?:(?:xn--)|[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]-?)*"
    r"[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]+)*"
    # TLD identifier
    r"(?:\.(?:(?:xn--[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]{2,})|"
    r"[a-z\u00a1-\uffff\U00010000-\U0010ffff]{2,}))"
    r")"
    # port number
    r"(?::\d{2,5})?"
    # resource path
    r"(?:/[-a-z\u00a1-\uffff\U00010000-\U0010ffff0-9._~%!$&'()*+,;=:@/]*)?"
    # query string
    r"(?:\?\S*)?"
    # fragment
    r"(?:#\S*)?"
    r"$",
    re.UNICODE | re.IGNORECASE
)


def url(val: str) -> bool:

    global REGEX_URL

    return True if REGEX_URL.match(val) else False


####################
# mac地址验证

REGEX_MAC_ADDR: re.Pattern = re.compile(r'^(?:[0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$')


def mac_addr(val: str) -> bool:

    global REGEX_MAC_ADDR

    return True if REGEX_MAC_ADDR.match(val) else False


####################
# IP地址验证

def ipv4(val: str) -> bool:

    groups = val.split(r'.')

    if len(groups) != 4 or any(not x.isdigit() for x in groups):
        return False

    return all(0 <= int(part) < 256 for part in groups)


def ipv4_cidr(val: str) -> bool:

    try:
        prefix, suffix = val.split(r'/', 2)
    except ValueError:
        return False

    if not ipv4(prefix) or not suffix.isdigit():
        return False

    return 0 <= int(suffix) <= 32


def ipv6(val: str) -> bool:

    ipv6_groups = val.split(r':')

    if len(ipv6_groups) == 1:
        return False

    ipv4_groups = ipv6_groups[-1].split(r'.')

    if len(ipv4_groups) > 1:

        if not ipv4(ipv6_groups[-1]):
            return False

        ipv6_groups = ipv6_groups[:-1]

    else:

        ipv4_groups = []

    max_groups = 6 if ipv4_groups else 8

    if len(ipv6_groups) > max_groups:
        return False

    count_blank = 0

    for part in ipv6_groups:

        if not part:
            count_blank += 1
            continue

        try:
            num = int(part, 16)
        except ValueError:
            return False
        else:
            if not 0 <= num <= 65536:
                return False

    if count_blank < 2:
        return True
    elif count_blank == 2 and not ipv6_groups[0] and not ipv6_groups[1]:
        return True

    return False


def ipv6_cidr(val: str) -> bool:

    try:
        prefix, suffix = val.split(r'/', 2)
    except ValueError:
        return False

    if not ipv6(prefix) or not suffix.isdigit():
        return False

    return 0 <= int(suffix) <= 128



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import time
import typing
import logging

import httpx

from abc import abstractmethod

from ...extend.base import Utils
from ...extend.struct import RoundRobinSimple
from ...extend.interface import TaskInterface
from ...extend.asyncio.future import Thread


DEFAULT_GROUP_NAME = r'DEFAULT_GROUP'
DEFAULT_PROTOCOL = r'http'

PROTOCOL_TYPE = typing.Literal[r'http', r'https']

DEFAULT_REQUEST_TIMEOUT = 10
DEFAULT_PULLING_TIMEOUT = 30
DEFAULT_WEIGHT = 1

WORD_SEPARATOR = u'\x02'
LINE_SEPARATOR = u'\x01'


logging.getLogger(r'httpx').setLevel(r'ERROR')
logging.getLogger(r'httpx').propagate = False


class _NacosInterface(TaskInterface):

    def __init__(
            self, servers: str, username: str, password: str,
            protocol: str, endpoint: str, group: str, namespace: str,
            request_timeout: float
    ):

        self._task: Thread = Thread(target=self._do_task)

        self._http_client: httpx.Client = httpx.Client(timeout=httpx.Timeout(request_timeout))

        self._servers: RoundRobinSimple[str] = RoundRobinSimple(
            Utils.split_str(servers, r',')
        )

        self._username: str = username
        self._password: str = password

        self._access_token: str = r''
        self._access_expires: int = 0

        self._protocol: str = protocol
        self._endpoint: str = endpoint
        self._group: str = group
        self._namespace: str = namespace

        self._request_timeout: float = request_timeout

    def _get_url_prefix(self) -> str:
        return f'{self._protocol}://{self._servers.get()}{self._endpoint}/nacos/v1'

    def _get_access_token(self) -> typing.Optional[str]:

        if not self._username or not self._password:
            return

        now_time = Utils.timestamp()

        if self._access_expires > now_time:
            return self._access_token

        server = self._servers.get()

        try:

            resp = self._http_client.post(
                f'{self._protocol}://{server}/nacos/v1/auth/login',
                data={r'username': self._username, r'password': self._password}
            )

            if resp.status_code == 200:

                content = resp.json()

                self._access_token = content[r'accessToken']
                self._access_expires = now_time + content[r'tokenTtl']

                Utils.log.info(f'nacos refresh access token expires: {Utils.stamp2time(self._access_expires)}')

            else:

                raise Exception(f'status code {resp.status_code}')

        except Exception as err:
            Utils.log.error(f'nacos {server} login error: {str(err)}')

        return self._access_token

    @abstractmethod
    def _do_task(self):
        """
        执行后台任务
        """

    def start(self):
        self._task.start()

    def stop(self):
        self._task.stop()
        self._http_client.close()

    def is_running(self) -> bool:
        return self._task.is_alive()


class NacosConfig(_NacosInterface):

    def __init__(
            self, listener: typing.Callable, servers: str, data_id: str, *,
            username: str = r'', password: str = r'',
            protocol: PROTOCOL_TYPE = DEFAULT_PROTOCOL, endpoint: str = r'',
            group: str = DEFAULT_GROUP_NAME, namespace: str = r'',
            request_timeout: float = DEFAULT_REQUEST_TIMEOUT, pulling_timeout: float = DEFAULT_PULLING_TIMEOUT
    ):

        super().__init__(servers, username, password, protocol, endpoint, group, namespace, request_timeout)

        self._listener: typing.Callable = listener

        self._data_id: str = data_id

        self._content: typing.Optional[str] = None
        self._content_has: typing.Optional[str] = None

        self._pulling_timeout: float = pulling_timeout

        self._cache_path: str = f'./.cache/nacos/{self._namespace}/{self._group}'

        self._flush()

    def _flush(self):

        self._content, self._content_hash = self.get_config()
        self._listener(self._content)

        Utils.log.info(f'nacos flush config: {self._content_hash}')

    def _write_cache(self, content: str):

        if not content:
            return

        if not os.path.exists(self._cache_path):
            os.makedirs(self._cache_path)

        with open(f'{self._cache_path}/{self._data_id}', 'wb') as file:
            file.write(content.encode())

    def _read_cache(self) -> typing.Optional[str]:

        if not os.path.exists(f'{self._cache_path}/{self._data_id}'):
            return None

        Utils.log.warning(f'nacos config used cache: {self._data_id}')

        with open(f'{self._cache_path}/{self._data_id}', 'rb') as file:
            return file.read().decode()

    def _do_task(self):

        while not self._task.is_stopped():

            url_prefix = self._get_url_prefix()

            payload = {
                r'Listening-Configs': WORD_SEPARATOR.join(
                    [
                        self._data_id,
                        self._group,
                        self._content_hash,
                        self._namespace,
                    ]
                ) + LINE_SEPARATOR,
            }

            params = {}

            if access_token := self._get_access_token():
                params[r'accessToken'] = access_token

            headers = {
                r'Long-Pulling-Timeout': str(self._pulling_timeout * 1000),
            }

            try:

                resp = self._http_client.post(
                    f'{url_prefix}/cs/configs/listener',
                    data=payload, params=params, headers=headers,
                    timeout=httpx.Timeout(self._request_timeout + self._pulling_timeout)
                )

                if resp.status_code == 200:
                    if resp.text:
                        Utils.log.info(f'nacos config pulling: {resp.text.strip()}')
                        self._flush()
                else:
                    raise Exception(r'nacos config pulling error')

            except httpx.TimeoutException:
                Utils.log.warning(f'nacos {url_prefix} config pulling timeout')
            except Exception as err:
                Utils.log.error(f'nacos {url_prefix} config pulling error: {str(err)}')
                time.sleep(self._request_timeout)

    def get_config(self) -> typing.Tuple[str, str]:

        content = content_hash = None

        url_prefix = self._get_url_prefix()

        params = {
            r'dataId': self._data_id,
            r'group': self._group,
        }

        if self._namespace:
            params[r'tenant'] = self._namespace

        if access_token := self._get_access_token():
            params[r'accessToken'] = access_token

        try:

            resp = self._http_client.get(f'{url_prefix}/cs/configs', params=params)

            if resp.status_code == 200:
                content = resp.text
                self._write_cache(content)
            else:
                raise Exception(f'status code {resp.status_code}')

        except Exception as err:
            Utils.log.error(f'nacos {url_prefix} get config error: {str(err)}')
            content = self._read_cache()

        if content is not None:
            content_hash = Utils.md5(content)

        return content, content_hash


class NacosInstanceRegister(_NacosInterface):

    def __init__(
            self, servers: str, service_name: str, service_ip: str, service_port: int, heartbeat_interval: int = 10, *,
            username: str = r'', password: str = r'',
            protocol: PROTOCOL_TYPE = DEFAULT_PROTOCOL, endpoint: str = r'',
            group: str = DEFAULT_GROUP_NAME, namespace: str = r'',
            cluster: typing.Optional[str] = None, weight: typing.Optional[int] = None,
            request_timeout: float = DEFAULT_REQUEST_TIMEOUT
    ):

        super().__init__(servers, username, password, protocol, endpoint, group, namespace, request_timeout)

        self._service_name: str = service_name
        self._service_ip: str = service_ip
        self._service_port: int = service_port

        self._cluster: str = cluster
        self._weight: int = weight if weight is not None else DEFAULT_WEIGHT

        self._heartbeat_interval: int = heartbeat_interval

    def _do_task(self):

        while not self._task.is_stopped():

            url_prefix = self._get_url_prefix()

            payload = {
                r'serviceName': self._service_name,
                r'namespaceId': self._namespace,
                r'groupName': self._group,
                r'beat': Utils.json_encode(
                    {
                        r'serviceName': self._service_name,
                        r'ip': self._service_ip,
                        r'port': str(self._service_port),
                        r'weight': self._weight,
                        r'ephemeral': True,
                    }
                ),
            }

            params = {}

            if access_token := self._get_access_token():
                params[r'accessToken'] = access_token

            try:

                resp = self._http_client.put(f'{url_prefix}/ns/instance/beat', data=payload, params=params)

                if resp.status_code != 200:
                    raise Exception(f'status code {resp.status_code}')

            except Exception as err:
                Utils.log.error(f'nacos {url_prefix} instance beat error: {str(err)}')

            time.sleep(self._heartbeat_interval)

    def start(self):

        url_prefix = self._get_url_prefix()

        payload = {
            r'serviceName': self._service_name,
            r'ip': self._service_ip,
            r'port': self._service_port,
            r'namespaceId': self._namespace,
            r'weight': self._weight,
            r'enabled': True,
            r'healthy': True,
            r'groupName': self._group,
            r'ephemeral': True,
        }

        if self._cluster:
            payload[r'clusterName'] = self._cluster

        params = {}

        if access_token := self._get_access_token():
            params[r'accessToken'] = access_token

        try:

            resp = self._http_client.post(f'{url_prefix}/ns/instance', data=payload, params=params)

            if resp.status_code == 200:
                self._task.start()
                Utils.log.info(f'nacos instance register: {payload}')
            else:
                raise Exception(f'status code {resp.status_code}')

        except Exception as err:
            Utils.log.error(f'nacos {url_prefix} instance register error: {str(err)}')


class NacosInstanceQuery(_NacosInterface):

    def __init__(
            self, servers: str, service_name: str, listener_interval: int = 10, *,
            listener_callback: typing.Optional[typing.Callable] = None,
            username: str = r'', password: str = r'',
            protocol: PROTOCOL_TYPE = DEFAULT_PROTOCOL, endpoint: str = r'',
            group: str = DEFAULT_GROUP_NAME, namespace: str = r'',
            cluster: typing.Optional[str] = None,
            request_timeout: float = DEFAULT_REQUEST_TIMEOUT
    ):

        super().__init__(servers, username, password, protocol, endpoint, group, namespace, request_timeout)

        self._service_name: str = service_name

        self._cluster: typing.Optional[str] = cluster

        self._content: typing.Optional[typing.List[typing.Dict]] = None
        self._content_hash: typing.Optional[str] = None

        self._listener_interval: int = listener_interval
        self._listener_callback: typing.Optional[typing.Callable] = listener_callback

        self._content, self._content_hash = self._send_query()

    def _do_task(self):

        while not self._task.is_stopped():

            content, content_hash = self._send_query()

            if content_hash != self._content_hash:

                Utils.log.info(f'nacos instance query: {content}')

                self._content, self._content_hash = content, content_hash

                if self._listener_callback is not None:
                    self._listener_callback(self._content)

            time.sleep(self._listener_interval)

    def _send_query(self) -> typing.Tuple[typing.List[typing.Dict], str]:

        content = content_hash = None

        url_prefix = self._get_url_prefix()

        params = {
            r'serviceName': self._service_name,
            r'namespaceId': self._namespace,
            r'groupName': self._group,
            r'healthyOnly': True,
        }

        if self._cluster:
            params[r'clusterName'] = self._cluster

        if access_token := self._get_access_token():
            params[r'accessToken'] = access_token

        try:

            resp = self._http_client.get(f'{url_prefix}/ns/instance/list', params=params)

            if resp.status_code == 200:
                content = sorted(
                    [
                        {
                            r'ip': _item[r'ip'],
                            r'port': _item[r'port'],
                            r'weight': _item[r'weight'],
                            r'metadata': _item[r'metadata'],
                        } for _item in resp.json()[r'hosts']
                    ],
                    key=lambda x: f"{x['ip']}:{x['port']}"
                )
                content_hash = Utils.md5(Utils.json_encode(content))
            else:
                raise Exception(f'status code {resp.status_code}')

        except Exception as err:
            Utils.log.error(f'nacos {url_prefix} instance query error: {str(err)}')

        return content, content_hash

    def get_host(self) -> typing.Dict:

        if self._content:
            return Utils.rand_hit(self._content, lambda x: int(x[r'weight'] * 100))

    def get_hosts(self) -> typing.List[typing.Dict]:
        return self._content



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from hagworm.frame.fastapi.base import APIRouter, Request
from hagworm.frame.fastapi.response import ErrorResponse

from service import DataSource


router = APIRouter()


@router.get(r'/')
async def default(request: Request):

    return request.client_ip, DataSource().timestamp_ms


@router.get(r'/error')
async def error(request: Request):

    raise ErrorResponse(-1, request.client_ip, 400)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import typing
import socket
import asyncio
import functools

import grpc

from .server import GRPCServer, Router
from .server import GRPC_MODE_TYPE, UNARY_UNARY_MODE, UNARY_STREAM_MODE, STREAM_UNARY_MODE, STREAM_STREAM_MODE
from .client import GRPCClient

from ...extend.asyncio.base import Utils
from ...extend.asyncio.command import MainProcess, SubProcess


def copy_metadata(remote_context: grpc.aio.ServicerContext) -> grpc.aio.Metadata:

    metadata = grpc.aio.Metadata((r'remote-peer', remote_context.peer()))

    for _key, _value in remote_context.invocation_metadata():
        metadata.add(_key, _value)

    return metadata


class _BridgeUpstream:

    def __init__(
            self, remote_request: typing.AsyncIterator,
            server_stream_stub: typing.Union[grpc.aio.StreamUnaryCall, grpc.aio.StreamStreamCall]
    ):

        self.__remote_request: typing.AsyncIterator = remote_request

        self._upstream_stub: typing.Union[grpc.aio.StreamUnaryCall, grpc.aio.StreamStreamCall] = server_stream_stub
        self._upstream_task: asyncio.Task = asyncio.create_task(self._do_upstream_task())

    async def _do_upstream_task(self):

        try:
            async for _message in self.__remote_request:
                await self._upstream_stub.write(_message)
            await self._upstream_stub.done_writing()
        except Exception as err:
            Utils.log.exception(str(err))


class _BridgeDownstream:

    def __init__(
            self, remote_context: grpc.aio.ServicerContext,
            server_stream_stub: typing.Union[grpc.aio.UnaryStreamCall, grpc.aio.StreamStreamCall]
    ):

        self.__remote_context: grpc.aio.ServicerContext = remote_context

        self._downstream_stub: typing.Union[grpc.aio.UnaryStreamCall, grpc.aio.StreamStreamCall] = server_stream_stub
        self._downstream_task: asyncio.Task = asyncio.create_task(self._do_downstream_task())

    async def _do_downstream_task(self):

        remote_peer = self.__remote_context.peer()

        Utils.log.info(f'grpc bridge open stream: {remote_peer}')

        try:
            async for _message in self._downstream_stub:
                await self.__remote_context.write(_message)
        except grpc.aio.UsageError:
            pass
        except Exception as err:
            Utils.log.exception(str(err))
        finally:
            if self._downstream_stub is not None:
                self._downstream_stub.cancel()

        Utils.log.info(f'grpc bridge close stream: {remote_peer}')


class BridgeUnaryStream(_BridgeDownstream):

    def __init__(
            self, server_bridge: GRPCClient, server_method: str,
            remote_request: typing.Union[str, bytes], remote_context: grpc.aio.ServicerContext
    ):

        super().__init__(
            remote_context,
            server_bridge.unary_stream(
                server_method, remote_request,
                metadata=copy_metadata(remote_context)
            )
        )

    async def join(self) -> typing.Any:
        return await self._downstream_task


class BridgeStreamUnary(_BridgeUpstream):

    def __init__(
            self, server_bridge: GRPCClient, server_method: str,
            remote_request: typing.AsyncIterator, remote_context: grpc.aio.ServicerContext
    ):

        super().__init__(
            remote_request,
            server_bridge.stream_unary(
                server_method,
                metadata=copy_metadata(remote_context)
            )
        )

    async def join(self) -> typing.Any:
        return await self._upstream_stub


class BridgeStreamStream(_BridgeUpstream, _BridgeDownstream):

    def __init__(
            self, server_bridge: GRPCClient, server_method: str,
            remote_request: typing.AsyncIterator, remote_context: grpc.aio.ServicerContext
    ):

        server_stream_stub = server_bridge.stream_stream(
            server_method,
            metadata=copy_metadata(remote_context)
        )

        _BridgeUpstream.__init__(self, remote_request, server_stream_stub)
        _BridgeDownstream.__init__(self, remote_context, server_stream_stub)

    async def join(self) -> typing.Any:
        return await asyncio.wait([self._upstream_task, self._downstream_task])


class GRPCBridge:

    def __init__(
            self, address: typing.Union[str, typing.Tuple[str, int]], routers: typing.List[Router],
            grpc_family: int = socket.AF_INET, grpc_server_credentials: typing.Optional[grpc.ServerCredentials] = None
    ):

        self._routers: typing.List[Router] = routers

        self._grpc_client: typing.Optional[GRPCClient] = None
        self._grpc_server: typing.Optional[GRPCServer] = None

        self._grpc_address: typing.Union[str, typing.Tuple[str, int]] = address
        self._grpc_family: int = grpc_family
        self._grpc_server_credentials: typing.Optional[grpc.ServerCredentials] = grpc_server_credentials

    async def _unary_unary_handler(self, method: str, request: typing.Union[str, bytes], context: grpc.aio.ServicerContext) -> typing.Any:
        try:
            return await self._grpc_client.unary_unary(method, request, metadata=copy_metadata(context))
        except grpc.aio.AioRpcError as err:
            await context.abort(err.code(), err.details())

    async def _unary_stream_handler(self, method: str, request: typing.Union[str, bytes], context: grpc.aio.ServicerContext) -> typing.Any:
        try:
            return await BridgeUnaryStream(self._grpc_client, method, request, context).join()
        except grpc.aio.AioRpcError as err:
            await context.abort(err.code(), err.details())

    async def _stream_unary_handler(self, method: str, request: typing.AsyncIterator, context: grpc.aio.ServicerContext) -> typing.Any:
        try:
            return await BridgeStreamUnary(self._grpc_client, method, request, context).join()
        except grpc.aio.AioRpcError as err:
            await context.abort(err.code(), err.details())

    async def _stream_stream_handler(self, method: str, request: typing.AsyncIterator, context: grpc.aio.ServicerContext) -> typing.Any:
        try:
            return await BridgeStreamStream(self._grpc_client, method, request, context).join()
        except grpc.aio.AioRpcError as err:
            await context.abort(err.code(), err.details())

    def _get_handler(self, method: str, mode: GRPC_MODE_TYPE) -> grpc.RpcMethodHandler:

        handler = None

        if mode is UNARY_UNARY_MODE:
            handler = functools.partial(self._unary_unary_handler, method)
        elif mode is UNARY_STREAM_MODE:
            handler = functools.partial(self._unary_stream_handler, method)
        elif mode is STREAM_UNARY_MODE:
            handler = functools.partial(self._stream_unary_handler, method)
        elif mode is STREAM_STREAM_MODE:
            handler = functools.partial(self._stream_stream_handler, method)

        return mode(handler)

    def _init_handlers(self):

        generic_handlers = [
            grpc.method_handlers_generic_handler(
                r'_',
                {
                    r'register': STREAM_STREAM_MODE(self._register_worker),
                    r'ping': self._get_handler(r'/_/ping', UNARY_UNARY_MODE),
                }
            )
        ]

        for router in self._routers:
            generic_handlers.append(
                grpc.method_handlers_generic_handler(
                    router.service,
                    {
                        _alias: self._get_handler(f'/{router.service}/{_alias}', _mode)
                        for _alias, _func, _mode in router.handlers
                    }
                )
            )

        self._grpc_server.server.add_generic_rpc_handlers(generic_handlers)

    async def _register_worker(self, request: typing.AsyncIterator, context: grpc.aio.ServicerContext):

        target = dict(context.invocation_metadata()).get(r'worker_entry_point')

        if target:
            context.add_done_callback(lambda _: Utils.create_task(self._grpc_client.clear_targets([target])))
        else:
            await context.abort(grpc.StatusCode.CANCELLED)

        await self._grpc_client.append_target(target)

        Utils.log.success(f'grpc bridge connected: {target}')

        with Utils.suppress(grpc.RpcError):
            async for _message in request:
                Utils.log.warning(f'grpc bridge received: {_message}')

    async def open(self):

        self._grpc_client: GRPCClient = GRPCClient(request_serializer=None, response_deserializer=None)
        self._grpc_server: GRPCServer = GRPCServer(request_deserializer=None, response_serializer=None)

        self._init_handlers()

        await self._grpc_server.start(
            self._grpc_address,
            family=self._grpc_family,
            server_credentials=self._grpc_server_credentials
        )

    async def close(self):

        await self._grpc_server.stop()
        await self._grpc_client.close()

    async def join(self):
        await self._grpc_server.wait()


class GRPCEntry:

    def __init__(self, bridge_address: typing.Union[str, typing.Tuple[str, int]], routers: typing.List[Router]):

        self._bridge_address: str = self._decode_address(bridge_address)
        self._server_address: str = f'/tmp/grpc_node_{Utils.getpid()}.sock'

        self._grpc_server: typing.Optional[GRPCServer] = None
        self._grpc_routers: typing.List[Router] = routers

        self._bridge_client: typing.Optional[GRPCClient] = None
        self._bridge_client_stub: typing.Optional[grpc.aio.StreamStreamCall] = None

    @staticmethod
    def _decode_address(bridge_address: typing.Union[str, typing.Tuple[str, int]]) -> str:

        if isinstance(bridge_address, str):
            bridge_address = Utils.split_str(bridge_address, r':')

        if bridge_address[0] == r'0.0.0.0':
            bridge_address =(r'localhost', bridge_address[1])

        return Utils.join_str(bridge_address, r':')

    async def start(self):

        if os.path.exists(self._server_address):
            os.remove(self._server_address)

        self._grpc_server = GRPCServer()

        for _router in self._grpc_routers:
            self._grpc_server.bind_router(_router)

        await self._grpc_server.start(self._server_address, family=socket.AF_UNIX, enable_ping=True)

        self._bridge_client = GRPCClient()
        await self._bridge_client.open([self._bridge_address])

        self._bridge_client_stub = self._bridge_client.stream_stream(
            r'/_/register',
            metadata=grpc.aio.Metadata((r'worker_entry_point', r'unix:' + self._server_address))
        )

    async def stop(self):
        await self._bridge_client.close()
        await self._grpc_server.stop()

    async def join(self):

        with Utils.suppress(grpc.RpcError):
            async for _message in self._bridge_client_stub:
                Utils.log.warning(f'grpc worker received: {_message}')

        await self._grpc_server.wait()


class GRPCMainProcess(MainProcess):

    def __init__(
            self, address: typing.Union[str, typing.Tuple[str, int]], routers: typing.List[Router],
            worker: typing.Type[r'GRPCWorker'], sub_process_num: int, *,
            grpc_family: int = socket.AF_INET, grpc_server_credentials: typing.Optional[grpc.ServerCredentials] = None,
            cpu_affinity: bool = False, join_timeout: int = 10
    ):

        super().__init__(
            worker.create, sub_process_num,
            cpu_affinity=cpu_affinity, join_timeout=join_timeout,
            bridge_address=address, routers=routers
        )

        self._grpc_bridge: GRPCBridge = GRPCBridge(address, routers, grpc_family, grpc_server_credentials)

    async def initialize(self):
        await self._grpc_bridge.open()

    async def release(self):
        await self._grpc_bridge.close()


class GRPCWorker(SubProcess):

    def __init__(self, process_num: int, bridge_address: typing.Union[str, typing.Tuple[str, int]], routers: typing.List[Router]):

        super().__init__(process_num)

        self._grpc_entry: GRPCEntry = GRPCEntry(bridge_address, routers)

    async def initialize(self):
        await self._grpc_entry.start()

    async def release(self):
        await self._grpc_entry.stop()

    async def _execute(self):
        await self._grpc_entry.join()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import typing

import json
import yaml

from ...extend.config import ConfigureBase

from .client import DEFAULT_PROTOCOL, PROTOCOL_TYPE, DEFAULT_GROUP_NAME, NacosConfig


class Configure(ConfigureBase):
    """配置类
    """

    __slots__ = [r'_format', r'_client']

    def __init__(
            self, servers: str, data_id: str, *,
            username: str = r'', password: str = r'',
            protocol: PROTOCOL_TYPE = DEFAULT_PROTOCOL, endpoint: str = r'',
            group: str = DEFAULT_GROUP_NAME, namespace: str = r'',
            data_format: typing.Literal[r'text', r'json', r'yaml'] = r'yaml'
    ):

        super().__init__()

        self._format: str = data_format

        self._client: typing.Optional[NacosConfig] = None

        if not servers and os.path.isfile(data_id):

            with open(data_id, 'r') as file:
                self._reload_config(file.read())

        else:

            self._client = NacosConfig(
                self._reload_config, servers, data_id,
                username=username, password=password,
                protocol=protocol, endpoint=endpoint, group=group, namespace=namespace
            )

    def _reload_config(self, content: str):

        self._clear_options()

        if self._format == r'text':
            self._parser.read_string(content)
        elif self._format == r'json':
            self._parser.read_dict(json.loads(content))
        elif self._format == r'yaml':
            self._parser.read_dict(yaml.load(content, yaml.Loader))

        self._load_options()

    def open(self):

        if self._client is not None:
            self._client.start()

    def close(self):

        if self._client is not None:
            self._client.stop()

        self._clear_options()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import asyncio

import aiormq
import aio_pika

from ...extend.asyncio.base import Utils, AsyncCirculatoryForSecond


class RabbitMQConsumer:
    """RabbitMQ消费者
    """

    def __init__(self, connection: aio_pika.RobustConnection, queue_name: str):

        self._connection: aio_pika.RobustConnection = connection
        self._channel: typing.Optional[aio_pika.abc.AbstractRobustChannel] = None

        self._queue: typing.Optional[aio_pika.abc.AbstractRobustQueue] = None
        self._queue_name: str = queue_name

    @property
    def channel(self) -> aio_pika.abc.AbstractRobustChannel:
        return self._channel

    @property
    def queue(self) -> aio_pika.abc.AbstractRobustQueue:
        return self._queue

    @property
    def queue_name(self) -> str:
        return self._queue_name

    async def open(
            self, *,
            consume_func: typing.Optional[typing.Callable] = None,
            consume_no_ack: bool = False,
            channel_qos_config: typing.Optional[typing.Dict] = None,
            queue_config: typing.Optional[typing.Dict] = None
    ):

        await self._connection.ready()

        if channel_qos_config is None:
            channel_qos_config = {r'prefetch_count': 1}
        elif r'prefetch_count' not in channel_qos_config:
            channel_qos_config[r'prefetch_count'] = 1

        if queue_config is None:
            queue_config = {}

        self._channel = await self._connection.channel()

        await self._channel.set_qos(**channel_qos_config)

        self._queue = await self._channel.declare_queue(self._queue_name, **queue_config)

        if consume_func is not None:
            await self._queue.consume(consume_func, no_ack=consume_no_ack)

        Utils.log.info(f"rabbitmq consumer {self._queue_name} initialized: {channel_qos_config[r'prefetch_count']}")

    async def close(self, with_connection: bool = False):

        await self._channel.close()

        if with_connection is True:
            await self._connection.close()

    async def block_pull(
            self,
            consume_func: typing.Callable,
            consume_no_ack: bool = False,
            wait_time: int = 1
    ):

        while not self._connection.is_closed or self._connection.reconnecting:

            try:

                message = await self._queue.get(no_ack=consume_no_ack, fail=False)

                if message is None:
                    await Utils.sleep(wait_time)
                else:
                    await consume_func(message)

            except Exception as err:

                Utils.log.error(str(err))
                await Utils.sleep(wait_time)

        Utils.log.warning(f'rabbitmq consumer block pull exit: {consume_func.__name__}')


class RabbitMQConsumerForExchange(RabbitMQConsumer):
    """RabbitMQ注册到交换机的消费者
    """

    def __init__(self, connection: aio_pika.RobustConnection, queue_name: str, exchange_name: str):
        
        super().__init__(connection, queue_name)

        self._exchange: typing.Optional[aio_pika.abc.AbstractExchange] = None
        self._exchange_name: str = exchange_name

        self._exchange_bind_task: typing.Optional[asyncio.Task] = None

    @property
    def exchange(self) -> aio_pika.abc.AbstractExchange:
        return self._exchange

    @property
    def exchange_name(self) -> str:
        return self._exchange_name

    async def _bind_exchange(self, routing_key: typing.Optional[str] = None):

        async for _ in AsyncCirculatoryForSecond(interval=10):

            try:
                self._exchange = await self._channel.get_exchange(self._exchange_name)
                await self._queue.bind(self._exchange, routing_key)
                break
            except aiormq.exceptions.ChannelNotFoundEntity as err:
                Utils.log.error(f'rabbitmq {self._queue_name} failed to bind {self._exchange_name}: {str(err)}')

        Utils.log.info(f'rabbitmq {self._queue_name} successfully bound {self._exchange_name}')

    async def open(
            self, *,
            consume_func: typing.Optional[typing.Callable] = None,
            consume_no_ack: bool = False,
            channel_qos_config: typing.Optional[typing.Dict] = None,
            queue_config: typing.Optional[typing.Dict] = None,
            routing_key: typing.Optional[str] = None
    ):

        await super().open(
            consume_func=consume_func, consume_no_ack=consume_no_ack,
            channel_qos_config=channel_qos_config, queue_config=queue_config
        )

        self._exchange_bind_task = asyncio.Task(self._bind_exchange(routing_key))

    async def block_pull(
            self,
            consume_func: typing.Callable,
            consume_no_ack: bool = False,
            wait_time: int = 1
    ):

        await self._exchange_bind_task
        await super().block_pull(consume_func, consume_no_ack, wait_time)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import io
import os
import sys
import copy
import signal
import typing
import queue
import logging
import datetime

import loguru

from abc import ABCMeta, abstractmethod
from datetime import timedelta
from contextlib import contextmanager

# noinspection PyProtectedMember
from loguru._file_sink import FileSink

from elasticsearch import Elasticsearch, AsyncElasticsearch, ApiError as ESApiError, NotFoundError as ESNotFoundError, helpers as es_helpers
from confluent_kafka import Producer as KFKProducer
from filelock import FileLock

from .trace import get_trace_id
from .base import Utils
from .asyncio.future import Thread


class LogFileRotator:

    @classmethod
    def make(cls, _size: int = 500, _time: str = r'00:00'):

        return cls(_size, _time).should_rotate

    def __init__(self, _size: int, _time: str):

        _size = _size * (1024 ** 2)
        _time = Utils.split_int(_time, r':')

        now_time = Utils.today()

        self._size_limit: int = _size
        self._time_limit: datetime.datetime = now_time.replace(hour=_time[0], minute=_time[1])

        if now_time >= self._time_limit:
            self._time_limit += timedelta(days=1)

    def should_rotate(self, message: 'loguru.Message', file: io.FileIO):

        file.seek(0, 2)

        if file.tell() + len(message) > self._size_limit:
            return True

        if message.record[r'time'].timestamp() > self._time_limit.timestamp():
            self._time_limit += timedelta(days=1)
            return True

        return False


DEFAULT_LOG_FILE_ROTATOR = LogFileRotator.make()


class InterceptHandler(logging.Handler):
    """日志拦截器
    """

    def emit(self, record: logging.LogRecord):

        Utils.log.opt(
            depth=8,
            exception=record.exc_info
        ).log(
            record.levelname,
            record.getMessage()
        )


class _BaseSink(metaclass=ABCMeta):
    """基础日志投递基类
    """

    def __init__(self, buffer_maxsize: int = 0xffff, bulk_maxsize: int = 0x800, write_max_delay: int = 5):

        self._buffer: queue.Queue = queue.Queue(buffer_maxsize)

        self._bulk_maxsize: int = bulk_maxsize
        self._write_max_delay: int = write_max_delay

        self._task: Thread = Thread(target=self._do_task)
        self._task.start()

        signal.signal(signal.SIGINT, self.close)
        signal.signal(signal.SIGTERM, self.close)

    def write(self, message: 'loguru.Message'):

        try:

            if message.record[r'thread'].id == self._task.ident or not message.record[r'message'].strip():
                return

            log_extra = message.record[r'extra']

            if r'trace_id' not in log_extra:

                trace_id = get_trace_id()

                if trace_id is not None:
                    log_extra[r'trace_id'] = trace_id

            self._buffer.put_nowait(message)

        except queue.Full as _:

            if message.record[r'level'].no > logging.INFO:
                sys.stderr.write(str(message))

        except Exception as err:

            sys.stderr.write(f'{str(err)}\n')

    def close(self, *_):

        self._task.stop(10)

    def _do_task(self):

        messages = []

        while not self._task.is_stopped():

            try:

                for _idx in range(self._write_max_delay):

                    try:

                        messages.append(
                            self._buffer.get(block=True, timeout=1)
                        )

                        self._buffer.task_done()

                    except queue.Empty:

                        if self._task.is_stopped():
                            break

                    finally:

                        if len(messages) >= self._bulk_maxsize or (_idx + 1) >= self._write_max_delay:
                            break

                if messages:
                    self._write_logs(messages)
                    messages.clear()

            except Exception as err:

                sys.stderr.write(f'{str(err)}\n')

    @abstractmethod
    def _write_logs(self, logs: typing.List['loguru.Message']):
        """
        实际写入日志的操作
        """


class QueuedFileSink(_BaseSink, FileSink):
    """日志文件队列
    """

    def __init__(
            self, path: str, *,
            buffer_maxsize: int = 0xffff, bulk_maxsize: int = 0x800, write_max_delay: int = 5,
            **kwargs
    ):

        _BaseSink.__init__(self, buffer_maxsize, bulk_maxsize, write_max_delay)
        FileSink.__init__(self, path, **kwargs)

    def _write_logs(self, logs: typing.List['loguru.Message']):

        for _log in logs:
            FileSink.write(self, _log)


class KafkaSink(_BaseSink):
    """Kafka日志投递
    """

    def __init__(
            self, servers: typing.Any, topic: str, *,
            buffer_maxsize: int = 0xffff, bulk_maxsize: int = 0x800, write_max_delay: int = 5,
            **kwargs
    ):

        super().__init__(buffer_maxsize, bulk_maxsize, write_max_delay)

        kwargs[r'bootstrap.servers'] = servers

        self._producer: KFKProducer = KFKProducer(kwargs)

        self._topic: str = topic

    def _write_logs(self, logs: typing.List['loguru.Message']):

        for _log in logs:
            self._producer.produce(
                self._topic,
                Utils.json_encode(
                    {
                        r'extra': _log.record[r'extra'],
                        r'process': {
                            r'id': _log.record[r'process'].id,
                            r'name': _log.record[r'process'].name,
                        },
                        r'thread': {
                            r'id': _log.record[r'thread'].id,
                            r'name': _log.record[r'thread'].name,
                        },
                        r'level': {
                            r'no': _log.record[r'level'].no,
                            r'name': _log.record[r'level'].name,
                        },
                        r'module': f"{_log.record[r'name']}:{_log.record[r'function']}:{_log.record[r'line']}",
                        r'message': _log.record[r'message'],
                        r'timestamp': int(_log.record[r'time'].timestamp() * 1000),
                    }
                ),
            )

            self._producer.poll(0)

        self._producer.flush()


class ElasticsearchDataStreamUtil:

    def __init__(
            self, elasticsearch: typing.Union[Elasticsearch, AsyncElasticsearch],
            stream_name: str, properties: typing.Dict, *,
            timestamp_field: str = r'@timestamp', timestamp_order: str = r'desc',
            rollover_max_age: str = r'7d', rollover_max_primary_shard_size: str = r'50gb', delete_min_age: str = r'30d',
            refresh_interval: str = r'5s', number_of_replicas=0
    ):

        self._elasticsearch: typing.Union[Elasticsearch, AsyncElasticsearch] = elasticsearch

        self._stream_name: str = stream_name
        self._policy_name: str = f'{stream_name}-ilm-policy'

        self._properties: typing.Dict = copy.deepcopy(properties)
        self._properties[r'@timestamp'] = {r'type': r'date'}

        self._rollover_max_age: str = rollover_max_age
        self._rollover_max_primary_shard_size: str = rollover_max_primary_shard_size

        self._delete_min_age: str = delete_min_age
        self._refresh_interval: str = refresh_interval
        self._number_of_replicas: int = number_of_replicas

        self._timestamp_field: str = timestamp_field
        self._timestamp_order: str = timestamp_order

    @contextmanager
    def _file_lock(self):

        if not os.path.exists(r'./.cache'):
            os.makedirs(r'./.cache')

        with FileLock(f'./.cache/log.{self._stream_name}.lock'):
            yield

    def initialize(self):

        with self._file_lock():
            self.create_lifecycle()
            self.create_index_template()

    def create_lifecycle(self):

        try:
            self._elasticsearch.ilm.get_lifecycle(name=self._policy_name)
        except ESNotFoundError as _:
            self._create_lifecycle()
        except ESApiError as err:
            sys.stderr.write(f'create lifecycle error: {str(err.info)}\n')

    def _create_lifecycle(self) -> typing.Any:

        policy = {
            r'phases': {
                r'hot': {
                    r'actions': {
                        r'rollover': {
                            r'max_age': self._rollover_max_age,
                            r'max_primary_shard_size': self._rollover_max_primary_shard_size,
                        },
                        r'set_priority': {
                            r'priority': 100,
                        }
                    },
                    r'min_age': r'0ms',
                },
                r'delete': {
                    r'actions': {
                        r'delete': {}
                    },
                    r'min_age': self._delete_min_age,
                },
            },
        }

        return self._elasticsearch.ilm.put_lifecycle(name=self._policy_name, policy=policy)

    def create_index_template(self):

        try:
            self._elasticsearch.indices.get_index_template(name=self._stream_name)
        except ESNotFoundError as _:
            self._create_index_template()
        except ESApiError as err:
            sys.stderr.write(f'{str(err.info)}\n')

    def _create_index_template(self) -> typing.Any:

        mappings = {
            r'dynamic': r'strict',
            r'properties': self._properties,
        }

        template = {
            r'settings': {
                r'index': {
                    r'lifecycle': {
                        r'name': self._policy_name,
                    },
                    r'refresh_interval': self._refresh_interval,
                    r'number_of_replicas': self._number_of_replicas,
                    r'sort': {
                        r'field': self._timestamp_field,
                        r'order': self._timestamp_order,
                    }
                },
            },
            r'mappings': mappings,
        }

        return self._elasticsearch.indices.put_index_template(
            name=self._stream_name,
            template=template,
            index_patterns=[f'{self._stream_name}*'],
            data_stream={},
        )


class AsyncElasticsearchDataStreamUtil(ElasticsearchDataStreamUtil):

    async def initialize(self):

        with self._file_lock():
            await self.create_lifecycle()
            await self.create_index_template()

    async def create_lifecycle(self):

        try:
            await self._elasticsearch.ilm.get_lifecycle(name=self._policy_name)
        except ESNotFoundError as _:
            await self._create_lifecycle()
        except ESApiError as err:
            sys.stderr.write(f'{str(err.info)}\n')

    async def create_index_template(self):

        try:
            await self._elasticsearch.indices.get_index_template(name=self._stream_name)
        except ESNotFoundError as _:
            await self._create_index_template()
        except ESApiError as err:
            sys.stderr.write(f'{str(err.info)}\n')


class ElasticsearchSink(_BaseSink):
    """Elasticsearch日志投递
    """

    def __init__(
            self, hosts: typing.Any, stream_name: str, *,
            buffer_maxsize: int = 0xffff, bulk_maxsize: int = 0x800, write_max_delay: int = 5,
            rollover_max_age: str = r'7d', rollover_max_primary_shard_size: str = r'50gb', delete_min_age: str = r'30d',
            refresh_interval: str = r'5s', number_of_replicas=0,
            **kwargs):

        super().__init__(buffer_maxsize, bulk_maxsize, write_max_delay)

        self._stream_name: str = stream_name

        self._elasticsearch: Elasticsearch = Elasticsearch(hosts, **kwargs)

        properties = {
            r'extra': {
                r'type': r'flattened',
            },
            r'process': {
                r'properties': {
                    r'id': {
                        r'type': r'keyword',
                    },
                    r'name': {
                        r'type': r'keyword',
                        r'ignore_above': 64,
                    },
                },
            },
            r'thread': {
                r'properties': {
                    r'id': {
                        r'type': r'keyword',
                    },
                    r'name': {
                        r'type': r'keyword',
                        r'ignore_above': 64,
                    },
                },
            },
            r'level': {
                r'properties': {
                    r'no': {
                        r'type': r'integer',
                    },
                    r'name': {
                        r'type': r'keyword',
                        r'ignore_above': 64,
                    },
                },
            },
            r'module': {
                r'type': r'text',
                r'norms': False,
            },
            r'message': {
                r'type': r'text',
                r'norms': False,
            },
            r'@timestamp': {
                r'type': r'date',
            }
        }

        ElasticsearchDataStreamUtil(
            self._elasticsearch, stream_name, properties,
            timestamp_field=r'@timestamp',
            timestamp_order=r'desc',
            rollover_max_age=rollover_max_age,
            rollover_max_primary_shard_size=rollover_max_primary_shard_size,
            delete_min_age=delete_min_age,
            refresh_interval=refresh_interval,
            number_of_replicas=number_of_replicas,
        ).initialize()

    def _write_logs(self, logs: typing.List['loguru.Message']):

        es_helpers.bulk(
            self._elasticsearch,
            actions=[
                {
                    r'_op_type': r'create',
                    r'_index': self._stream_name,
                    r'extra': _log.record[r'extra'],
                    r'process': {
                        r'id': _log.record[r'process'].id,
                        r'name': _log.record[r'process'].name,
                    },
                    r'thread': {
                        r'id': _log.record[r'thread'].id,
                        r'name': _log.record[r'thread'].name,
                    },
                    r'level': {
                        r'no': _log.record[r'level'].no,
                        r'name': _log.record[r'level'].name,
                    },
                    r'module': f"{_log.record[r'name']}:{_log.record[r'function']}:{_log.record[r'line']}",
                    r'message': _log.record[r'message'],
                    r'@timestamp': int(_log.record[r'time'].timestamp() * 1000),
                }
                for _log in logs
            ]
        )


DEFAULT_LOG_FILE_NAME = r'runtime_{time}.log'


def init_logger(
        level: str, *, handler: typing.Optional[logging.Handler] = None,
        file_path: typing.Optional[str] = None, file_name: str = DEFAULT_LOG_FILE_NAME,
        file_rotation: typing.Callable = DEFAULT_LOG_FILE_ROTATOR, file_retention: int = 0xff,
        extra: typing.Optional[typing.Dict] = None, enqueue: bool = False, debug: bool = False
):

    level = level.upper()

    Utils.log.remove()

    if extra is not None:

        extra = {_key: _val for _key, _val in extra.items() if _val is not None}

        if extra:
            Utils.log.configure(extra=extra)

    if handler or file_path:

        if handler:
            Utils.log.add(
                handler,
                level=level,
                enqueue=enqueue,
                backtrace=debug
            )

        if file_path:

            _file_name, _file_ext_name = os.path.splitext(file_name)

            Utils.log.add(
                QueuedFileSink(
                    Utils.path.join(file_path, _file_name + '.pid-' + str(Utils.getpid()) + _file_ext_name),
                    rotation=file_rotation,
                    retention=file_retention
                ),
                level=level,
                enqueue=enqueue,
                backtrace=debug
            )

    else:

        Utils.log.add(
            sys.stderr,
            level=level,
            enqueue=enqueue,
            backtrace=debug
        )

    logger = logging.getLogger()
    logger.setLevel(level)
    logger.addHandler(InterceptHandler())



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import time
import pandas

from datetime import datetime
from ntplib import NTPClient as _NTPClient
from apscheduler.schedulers.background import BackgroundScheduler

from hagworm.extend.base import Utils
from hagworm.extend.error import catch_error
from hagworm.extend.asyncio.task import DEFAULT_SCHEDULER_CONFIG


class NTPClient:
    """异步NTP客户端类
    """

    def __init__(
            self, host: str = r'time.windows.com', *,
            version: int = 4, port: str = r'ntp', timeout: int = 5,
            interval: int = 3600, sampling: int = 5
    ):

        self._client: _NTPClient = _NTPClient()
        self._offset: float = 0
        self._sampling: int = sampling

        self._scheduler: BackgroundScheduler = BackgroundScheduler(**DEFAULT_SCHEDULER_CONFIG)

        self._scheduler.add_job(
            self._do_task, 'interval',
            (host, version, port, timeout),
            seconds=interval, next_run_time=datetime.now()
        )

    def start(self):
        if not self._scheduler.running:
            self._scheduler.start()

    def stop(self):
        if self._scheduler.running:
            self._scheduler.shutdown()

    def _do_task(self, host: str, version: int, port: str, timeout: int):

        with catch_error():

            samples = []

            # 多次采样取中位数，减少抖动影响
            for _ in range(self._sampling):
                try:
                    _response = self._client.request(host, version, port, timeout)
                    samples.append(_response.offset)
                except Exception as err:
                    Utils.log.error(f'NTP server {host} request error: {str(err)}')

            if samples:
                self._offset = pandas.Series(samples).median()
                Utils.log.info(f'NTP server {host} offset median {self._offset} samples: {samples}')
            else:
                Utils.log.error(f'NTP server {host} not available, timestamp uncalibrated')

    @property
    def offset(self) -> float:
        return self._offset

    @property
    def timestamp(self) -> float:
        return time.time() + self._offset

    @property
    def timestamp_ms(self) -> int:
        return round((time.time() + self._offset) * 1000)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import typing
import threading
import warnings
import traceback


class SingletonMetaclass(type):
    """单例的元类实现
    """

    def __init__(cls, __name: str, __bases: tuple[type, ...], __dict: dict[str, typing.Any]):

        super().__init__(__name, __bases, __dict)

        cls._instance: typing.Any = None
        cls._process_id: typing.Optional[int] = None

    def __call__(cls, *args, **kwargs) -> typing.Any:

        # noinspection PyBroadException
        try:

            if cls._instance is not None:

                cls.check_process_id()

                return cls._instance

            else:

                cls._instance = super().__call__(*args, **kwargs)
                cls._process_id = os.getpid()

                return cls._instance

        except Exception as _:

            traceback.print_exc()

    def check_process_id(cls) -> bool:

        if cls._process_id == os.getpid():
            return True
        else:
            warnings.warn(f'Memory replication may cause system resource conflicts: {cls.__qualname__}')
            return False


class Singleton(metaclass=SingletonMetaclass):
    """单例基类
    """
    pass


class SafeSingletonMetaclass(SingletonMetaclass):
    """线程安全的单例的元类实现
    """

    def __init__(cls, __name: str, __bases: tuple[type, ...], __dict: dict[str, typing.Any]):

        super().__init__(__name, __bases, __dict)

        cls._lock: threading.Lock = threading.Lock()

    def __call__(cls, *args, **kwargs) -> typing.Any:

        with cls._lock:
            return super().__call__(*args, **kwargs)


class SafeSingleton(metaclass=SafeSingletonMetaclass):
    """线程安全的单例基类
    """
    pass



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import sys

os.chdir(os.path.dirname(__file__))
sys.path.insert(0, os.path.abspath(r'../'))

from hagworm.extend.asyncio.base import Utils
from hagworm.frame.stress_tests import Launcher, RunnerAbstract, TimerMS


class Runner(RunnerAbstract):

    async def _execute(self):

        for index in range(5):

            for _ in range(5):

                with TimerMS() as timer:
                    await Utils.sleep(Utils.randint(10, 99) / 1000)

                if Utils.rand_hit([True, False], [50, 50]):
                    await self.success(f'Test{index}', timer.value)
                else:
                    await self.failure(f'Test{index}', timer.value)


if __name__ == r'__main__':

    Launcher(Runner.create, 5).run()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import traceback

from loguru import logger
from contextlib import contextmanager


class Ignore(Exception):
    """可忽略的异常

    用于with语句块跳出，或者需要跳出多层逻辑的情况

    """

    def __init__(self, *args, layers=1):

        super().__init__(*args)

        self._layers = layers

        info = str(self)

        if info:
            logger.warning(info)

    def throw(self):

        if self._layers > 0:
            self._layers -= 1

        return self._layers != 0


@contextmanager
def catch_warning():
    """异常捕获，打印warning级日志

    通过with语句捕获异常，代码更清晰，还可以使用Ignore异常安全的跳出with代码块

    """

    # noinspection PyBroadException
    try:
        yield
    except Ignore as err:
        if err.throw():
            raise err
    except Exception as _:
        logger.warning(traceback.format_exc())


@contextmanager
def catch_error():
    """异常捕获，打印error级日志

    通过with语句捕获异常，代码更清晰，还可以使用Ignore异常安全的跳出with代码块

    """

    # noinspection PyBroadException
    try:
        yield
    except Ignore as err:
        if err.throw():
            raise err
    except Exception as _:
        logger.error(traceback.format_exc())



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import setuptools

from hagworm import __version__


with open(r'README.md', encoding=r'utf8') as stream:
    long_description = stream.read()

setuptools.setup(
    name=r'hagworm',
    version=__version__,
    license=r'Apache License Version 2.0',
    platforms=[r'all'],
    author=r'Shaobo.Wang',
    author_email=r'wsb310@gmail.com',
    description=r'Network Development Suite',
    long_description=long_description,
    long_description_content_type=r'text/markdown',
    url=r'https://gitee.com/wsb310/hagworm',
    packages=setuptools.find_packages(),
    python_requires=r'>=3.10',
    install_requires=[
        r'Cython==3.0.10',
        r'APScheduler==3.10.4',
        r'ahocorasick-rs==0.22.0',
        r'Pillow==10.4.0',
        r'PyJWT==2.8.0',
        r'PyYAML==6.0.1',
        r'SQLAlchemy==2.0.31',
        r'coredis==4.17.0',
        r'asyncmy==0.2.9',
        r'aiosmtplib==3.0.1',
        r'aio-pika==9.4.2',
        r'cachetools==5.3.3',
        r'confluent-kafka==2.5.0',
        r'cryptography==42.0.8',
        r'elasticsearch==8.12.1',
        r'fastapi==0.111.0',
        r'filelock==3.15.4',
        r'gunicorn==22.0.0',
        r'grpcio==1.60.0',
        r'httpx==0.27.0',
        r'httptools==0.6.1',
        r'igraph==0.11.6',
        r'loguru==0.7.2',
        r'motor==3.5.1',
        r'msgpack==1.0.8',
        r'ntplib==0.4.0',
        r'psutil==6.0.0',
        r'pandas==2.2.2',
        r'pytest-asyncio==0.23.7',
        r'python-dateutil==2.8.2',
        r'python-stdnum==1.20',
        r'python-multipart==0.0.9',
        r'pymongo==4.8.0',
        r'qrcode==7.4.2',
        r'texttable==1.7.0',
        r'ujson==5.10.0',
        r'uvicorn[standard]==0.30.1',
        r'uvloop==0.19.0;sys_platform!="win32"',
        r'xmltodict==0.13.0',
    ],
    classifiers=[
        r'Programming Language :: Python :: 3.10',
        r'License :: OSI Approved :: Apache Software License',
        r'Operating System :: POSIX :: Linux',
    ],
)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from hagworm.frame.fastapi.base import APIRouter

from . import home


router = APIRouter(prefix=r'/demo')

router.include_router(home.router)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from hagworm.extend.trace import get_trace_id
from hagworm.extend.error import Ignore
from hagworm.extend.metaclass import Singleton
from hagworm.extend.asyncio.base import Utils
from hagworm.extend.asyncio.redis import RedisDelegate
from hagworm.extend.asyncio.mongo import MongoDelegate
from hagworm.extend.asyncio.mysql import MySQLDelegate
from hagworm.extend.ntp import NTPClient

from setting import Config


class DataSource(Singleton, RedisDelegate, MongoDelegate, MySQLDelegate):

    def __init__(self):

        # RedisDelegate.__init__(self)
        # MongoDelegate.__init__(self)
        # MySQLDelegate.__init__(self)

        self._ntp_client = NTPClient()

    @property
    def timestamp_ms(self) -> int:
        return self._ntp_client.timestamp_ms

    @classmethod
    async def initialize(cls):

        inst = cls()

        inst._ntp_client.start()

        # (await inst.init_redis_single(
        #     Config.RedisHost[0], Config.RedisHost[1], password=Config.RedisPasswd,
        #     min_connections=Config.RedisMinConn, max_connections=Config.RedisMaxConn
        # )).set_key_prefix(Config.RedisKeyPrefix)
        #
        # inst.init_mongo(
        #     Config.MongoHost, Config.MongoUser, Config.MongoPasswd,
        #     auth_source=Config.MongoAuth,
        #     min_pool_size=Config.MongoMinConn, max_pool_size=Config.MongoMaxConn, max_idle_time=3600
        # )
        #
        # if Config.MySqlMasterServer:
        #
        #     await inst.init_mysql_rw(
        #         Config.MySqlMasterServer[0], Config.MySqlMasterServer[1], Config.MySqlName,
        #         Config.MySqlUser, Config.MySqlPasswd,
        #         minsize=Config.MySqlMasterMinConn, maxsize=Config.MySqlMasterMaxConn,
        #         echo=Config.Debug, pool_recycle=21600
        #     )
        #
        # if Config.MySqlSlaveServer:
        #
        #     await inst.init_mysql_ro(
        #         Config.MySqlSlaveServer[0], Config.MySqlSlaveServer[1], Config.MySqlName,
        #         Config.MySqlUser, Config.MySqlPasswd,
        #         minsize=Config.MySqlSlaveMinConn, maxsize=Config.MySqlSlaveMaxConn,
        #         echo=Config.Debug, pool_recycle=21600
        #     )

    @classmethod
    async def release(cls):

        inst = cls()

        # await inst.close_mysql()
        #
        # inst.close_mongo_pool()
        #
        # await inst.close_redis_pool()

        inst._ntp_client.stop()


class ServiceBase(Singleton, Utils):

    def __init__(self):

        self._data_source = DataSource()

    @property
    def trace_id(self):

        return get_trace_id()

    def Break(self, *args, layers=1):

        raise Ignore(*args, layers=layers)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from abc import ABC, abstractmethod
from typing import Any, Union, Literal, Optional, Iterable, List, Dict, Tuple

from coredis.typing import Node
from coredis import Redis, RedisCluster, Connection
from coredis.recipes.locks import LuaLock as _LuaLock

from .base import Utils

from ..error import catch_error
from ..interface import AsyncContextManager
from .transaction import Transaction


class LuaLock(AsyncContextManager, _LuaLock):

    def __init__(
            self, client: Any, name: str,
            timeout: Optional[float] = None, sleep: float = 0.1,
            blocking: bool = True, blocking_timeout: Optional[float] = None
    ):
        _LuaLock.__init__(self, client, name, timeout, sleep, blocking, blocking_timeout)

    async def _context_release(self):
        with catch_error():
            await self.release()

    async def acquire(self) -> bool:
        if self.local.get() is None:
            return await super().acquire()
        else:
            return await super().extend(self.timeout)

    async def release(self):
        if self.local.get() is not None:
            await super().release()


class _PoolMixin(ABC):

    def __init__(self):

        self._name: str = Utils.uuid1()[:8]
        self._key_prefix: Optional[str] = None

    # noinspection PyUnresolvedReferences
    async def _init_connection(self):

        connections = []

        for _ in range(self.max_connections):
            connection = await self._get_connection()
            await connection.connect()
            connections.append(connection)

        for connection in connections:
            self.connection_pool.release(connection)

    @abstractmethod
    async def _get_connection(self) -> Connection:
        raise NotImplementedError()

    @staticmethod
    def decode(value: Union[bytes, Tuple[bytes, ...], List[bytes], Dict[bytes, bytes]]) -> Any:

        type_ = type(value)

        if type_ is bytes:
            return value.decode(r'utf-8')
        elif type_ is tuple:
            return tuple(
                _val.decode(r'utf-8') if isinstance(_val, bytes) else _val
                for _val in value
            )
        elif type_ is list:
            return [
                _val.decode(r'utf-8') if isinstance(_val, bytes) else _val
                for _val in value
            ]
        elif type_ is dict:
            return {
                _key.decode(r'utf-8'): _val.decode(r'utf-8') if isinstance(_val, bytes) else _val
                for _key, _val in value.items()
            }
        else:
            return value

    # noinspection PyUnresolvedReferences
    @property
    def max_connections(self) -> int:
        return self.connection_pool.max_connections

    def set_key_prefix(self, value: str):

        self._key_prefix = value

    def get_safe_key(self, key, *args, **kwargs) -> str:

        if self._key_prefix:
            _key = f'{self._key_prefix}:{key}'
        else:
            _key = key

        if args or kwargs:
            _key = f'{_key}:{Utils.params_sign(*args, **kwargs)}'

        return _key

    # noinspection PyUnresolvedReferences
    async def close(self):
        self.connection_pool.disconnect()

    def allocate_lock(
            self, name: str,
            timeout: Optional[float] = None, sleep: float = 0.1,
            blocking: bool = True, blocking_timeout: Optional[float] = None
    ) -> LuaLock:

        return LuaLock(self, name, timeout, sleep, blocking, blocking_timeout)

    # noinspection PyUnresolvedReferences
    async def get_obj(self, name: str) -> Any:

        result = await self.get(name)

        return Utils.pickle_loads(result) if result else result

    # noinspection PyUnresolvedReferences
    async def getset_obj(self, name: str, value: Any) -> Any:

        _value = Utils.pickle_dumps(value)

        result = await self.getset(name, _value)

        return Utils.pickle_loads(result) if result else result

    # noinspection PyUnresolvedReferences
    async def set_obj(self, name: str, value: Any, seconds: Optional[float] = None):

        _value = Utils.pickle_dumps(value)

        if seconds is None:
            return await self.set(name, _value)
        else:
            return await self.setex(name, _value, seconds)


class RedisPool(_PoolMixin, Redis):
    """StrictRedis连接管理
    """

    def __init__(
            self,
            host:str, port:int, db: int = 0,
            username: Optional[str] = None, password: Optional[str] = None,
            max_connections: int =32, max_idle_time: float = 43200, idle_check_interval: float = 1,
            protocol_version: Literal[2, 3] = 2,
            **kwargs
    ):

        Redis.__init__(
            self,
            host, port, db,
            username=username, password=password, max_connections=max_connections,
            max_idle_time=max_idle_time, idle_check_interval=idle_check_interval,
            protocol_version=protocol_version,
            **kwargs
        )

        _PoolMixin.__init__(self)

    async def _get_connection(self) -> Connection:
        return await self.connection_pool.get_connection()

    async def open(self):

        await Redis.initialize(self)

        await self._init_connection()

        config = self.connection_pool.connection_kwargs

        Utils.log.info(
            f"Redis Pool ({self._name}) [{config[r'host']}:{config[r'port']}] initialized: {self.max_connections}"
        )

        return self


class RedisClusterPool(_PoolMixin, RedisCluster):
    """StrictRedisCluster连接管理
    """

    def __init__(
            self,
            host: str, port: int, startup_nodes: Optional[Iterable[Node]] = None,
            username: Optional[str] = None, password: Optional[str] = None,
            max_connections: int = 32, max_idle_time: float = 43200, idle_check_interval: float = 1,
            protocol_version: Literal[2, 3] = 2,
            **kwargs
    ):

        RedisCluster.__init__(
            self,
            host, port, startup_nodes=startup_nodes,
            username=username, password=password, max_connections=max_connections,
            max_idle_time=max_idle_time, idle_check_interval=idle_check_interval,
            protocol_version=protocol_version,
            **kwargs
        )

        _PoolMixin.__init__(self)

    async def _get_connection(self) -> Connection:
        return await self.connection_pool.get_random_connection()

    async def open(self):

        await RedisCluster.initialize(self)

        await self._init_connection()

        nodes = self.connection_pool.nodes.nodes

        Utils.log.info(
            f"Redis Cluster Pool ({self._name}) {list(nodes.keys())} initialized: {self.max_connections}"
        )

        return self


class RedisDelegate:
    """Redis功能组件
    """

    def __init__(self):

        self._redis_pool: Optional[RedisPool, RedisClusterPool] = None

    @property
    def redis_pool(self) -> Union[RedisPool, RedisClusterPool]:

        return self._redis_pool

    async def init_redis_single(
            self,
            host:str, port:int, db: int = 0,
            username: Optional[str] = None, password: Optional[str] = None, *,
            max_connections: int =32, max_idle_time: float = 43200, idle_check_interval: float = 1,
            protocol_version: Literal[2, 3] = 2,
            **kwargs
    ):

        self._redis_pool = await RedisPool(
            host, port, db, username, password,
            max_connections, max_idle_time, idle_check_interval,
            protocol_version,
            **kwargs
        ).open()

        return self._redis_pool

    async def init_redis_cluster(
            self,
            host: str, port: int, startup_nodes: Optional[Iterable[Node]] = None,
            username: Optional[str] = None, password: Optional[str] = None, *,
            max_connections: int = 32, max_idle_time: float = 43200, idle_check_interval: float = 1,
            protocol_version: Literal[2, 3] = 2,
            **kwargs
    ):

        self._redis_pool = await RedisClusterPool(
            host, port, startup_nodes, username, password,
            max_connections, max_idle_time, idle_check_interval,
            protocol_version,
            **kwargs
        ).open()

        return self._redis_pool

    def set_redis_key_prefix(self, value: str):

        self._redis_pool.set_key_prefix(value)

    async def close_redis_pool(self):

        await self._redis_pool.close()

    def get_redis_client(self) -> Union[RedisPool, RedisClusterPool]:

        return self._redis_pool


class ShareCache(AsyncContextManager):
    """共享缓存，使用with进行上下文管理

    基于分布式锁实现的一个缓存共享逻辑，保证在分布式环境下，同一时刻业务逻辑只执行一次，其运行结果会通过缓存被共享

    """

    def __init__(
            self, redis_client: Union[RedisPool, RedisClusterPool], share_key: str,
            lock_timeout: float = 60, lock_blocking_timeout: float = 60
    ):

        self._redis_client: Union[RedisPool, RedisClusterPool] = redis_client
        self._share_key: str = redis_client.get_safe_key(share_key)

        self._lock: LuaLock = self._redis_client.allocate_lock(
            redis_client.get_safe_key(f'share_cache:{share_key}'),
            timeout=lock_timeout, blocking=True, blocking_timeout=lock_blocking_timeout
        )

        self.result = None

    async def _context_release(self):

        await self.release()

    async def get(self) -> Any:

        result = await self._redis_client.get_obj(self._share_key)

        if result is None:

            if await self._lock.acquire():
                result = await self._redis_client.get_obj(self._share_key)

        return result

    async def set(self, value: Any, expire: Optional[float] = None):

        return await self._redis_client.set_obj(self._share_key, value, expire)

    async def release(self):

        if self._lock:
            await self._lock.release()

        self._redis_client = self._lock = None


class PeriodCounter:

    def __init__(
            self, redis_client: Union[RedisPool, RedisClusterPool],
            key_prefix: str, time_slice: int,
            expire: Optional[float] = None
    ):

        self._redis_client: Union[RedisPool, RedisClusterPool] = redis_client

        self._key_prefix = key_prefix
        self._time_slice = time_slice

        self._expire = time_slice if expire is None else expire

    def _get_key(self, offset: int = 0) -> str:

        time_period = Utils.math.floor(Utils.timestamp() / self._time_slice) + offset

        return self._redis_client.get_safe_key(f'{self._key_prefix}:{time_period}')

    async def get(self, offset: int = 0):

        return await self._redis_client.get(
            self._get_key(offset)
        )

    async def _incr_by(self, key: str, val: int) -> int:

        async with await self._redis_client.pipeline(transaction=False) as pipeline:

            await pipeline.incrby(key, val)
            await pipeline.expire(key, max(self._time_slice, self._expire))

            res, _ = await pipeline.execute()

            return res

    async def _decr_by(self, key: str, val: int) -> int:

        async with await self._redis_client.pipeline(transaction=False) as pipeline:

            await pipeline.decrby(key, val)
            await pipeline.expire(key, max(self._time_slice, self._expire))

            res, _ = await pipeline.execute()

            return res

    async def incr(self, val: int = 1) -> int:

        return await self._incr_by(self._get_key(), val)

    async def incr_with_trx(self, val: int = 1) -> Tuple[int, Transaction]:

        _key = self._get_key()

        res = await self._incr_by(_key, val)

        if res is not None:
            trx = Transaction()
            trx.add_rollback_callback(self._decr_by, _key, val)
        else:
            trx = None

        return res, trx

    async def decr(self, val: int = 1) -> int:

        return await self._decr_by(self._get_key(), val)

    async def decr_with_trx(self, val: int = 1) -> Tuple[int, Transaction]:

        _key = self._get_key()

        res = await self._decr_by(_key, val)

        if res is not None:
            trx = Transaction()
            trx.add_rollback_callback(self._incr_by, _key, val)
        else:
            trx = None

        return res, trx



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import abc
import time
import math
import typing
import itertools
import pandas

from texttable import Texttable

from ..extend.base import Utils
from ..extend.asyncio.socket import recv_msg, DEFAULT_UNIX_SOCKET_ENDPOINT
from ..extend.asyncio.command import MainProcessWithIPC, SubProcessWithIPC


class TimerMS:

    __slots__ = [r'_start_time', r'_stop_time']

    def __init__(self, timer: typing.Optional[float] = None):
        self._start_time: float = time.time() if timer is None else timer
        self._stop_time: typing.Optional[float] = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.done()

    def done(self):
        self._stop_time = (time.time() - self._start_time) * 1000

    @property
    def value(self) -> float:
        return self._stop_time


class RunnerAbstract(SubProcessWithIPC):

    @abc.abstractmethod
    async def _execute(self):
        """
        实际测试的业务逻辑
        """

    async def success(self, name: str, resp_time: typing.Union[int, float]):

        await self._socket_client.send_msg(
            [name, True, resp_time]
        )

    async def failure(self, name: str, resp_time: typing.Union[int, float]):

        await self._socket_client.send_msg(
            [name, False, resp_time]
        )


class Launcher(MainProcessWithIPC):

    def __init__(
            self, target: typing.Callable, sub_process_num: int, *args,
            time_unit: str = r'ms', shard_table_bins: typing.List[float] = None,
            unix_socket_path: str = DEFAULT_UNIX_SOCKET_ENDPOINT, **kwargs
    ):

        super().__init__(target, sub_process_num, *args, unix_socket_path=unix_socket_path, **kwargs)

        self._data_frame: pandas.DataFrame = pandas.DataFrame(
            columns=[r'name', r'success', r'response_time']
        )

        self._time_unit: str = time_unit

        self._shard_table_bins: typing.List[float] = shard_table_bins if shard_table_bins else list(range(0, 100, 20))
        self._shard_table_bins.append(math.inf)

    async def _client_connected_cb(self, reader, writer):

        while True:

            request = await recv_msg(reader)

            if request:
                self._data_frame.loc[self._data_frame.size] = request
            else:
                break

    def _get_base_table(self) -> str:

        table = Texttable(max_width=0)

        table.header(
            [
                r'Event Name',
                r'Success Total',
                r'Failure Total',
                r'Success Ratio',
                r'Success Ave',
                r'Success Std',
                r'Success Min',
                r'Success Max',
            ]
        )

        for _name, _data_frame in self._data_frame.groupby(self._data_frame.name):

            _success_data_frame = _data_frame[_data_frame.apply(lambda x: x[r'success'], axis=1)]
            _failure_data_frame = _data_frame[_data_frame.apply(lambda x: not x[r'success'], axis=1)]

            table.add_row(
                [
                    _name,
                    _success_data_frame.shape[0],
                    _failure_data_frame.shape[0],
                    r'{:.2%}'.format(_success_data_frame.shape[0] / max(_data_frame.shape[0], 1)),
                    r'{:.3f}{}'.format(_success_data_frame[r'response_time'].mean(), self._time_unit),
                    r'{:.3f}'.format(_success_data_frame[r'response_time'].std()),
                    r'{:.3f}{}'.format(_success_data_frame[r'response_time'].min(), self._time_unit),
                    r'{:.3f}{}'.format(_success_data_frame[r'response_time'].max(), self._time_unit),
                ]
            )

        return table.draw()

    def _get_series_table(self) -> str:

        table = Texttable(max_width=0)

        labels = [f'{v1}-{v2}({self._time_unit})' for v1, v2 in itertools.pairwise(self._shard_table_bins)]

        table.header([r'Event Name'] + labels)

        for _name, _data_frame in self._data_frame[
            self._data_frame.apply(lambda x: x[r'success'], axis=1)
        ].groupby(self._data_frame.name):

            _series = pandas.cut(
                _data_frame[r'response_time'], self._shard_table_bins, labels=labels, retbins=True
            )[0].value_counts(sort=False, normalize=True)

            _rows = [_name]

            for _label in labels:
                _rows.append(
                    r'{:.2%}'.format(_series.get(_label, 0))
                )

            table.add_row(_rows)

        return table.draw()

    async def _execute(self):

        with TimerMS() as timer:
            await super()._execute()

        _time = max(round(timer.value / 1000), 1)
        _count = self._data_frame.shape[0]

        Utils.log.info(
            f'\nTotal: {_count}, Time: {_time}s, Qps: {round(_count / _time)}'
            f'\n{self._get_base_table()}\nSuccess Data Distribution\n{self._get_series_table()}\n'
        )
